"""A List of MySQL SQL keywords.

https://dev.mysql.com/doc/refman/8.0/en/keywords.html
"""

mysql_reserved_keywords = """ACCESSIBLE
ADD
ALL
ALTER
ANALYZE
AND
AS
ASC
ASENSITIVE
BEFORE
BETWEEN
BIGINT
BINARY
BLOB
BOTH
BY
CALL
CASCADE
CASE
CHANGE
CHAR
CHARACTER
CHECK
COLLATE
COLUMN
CONDITION
CONSTRAINT
CONTINUE
CONVERT
CREATE
CROSS
CUME_DIST
CURRENT_DATE
CURRENT_TIME
CURRENT_TIMESTAMP
CURRENT_USER
CURSOR
DATABASE
DATABASES
DAY_HOUR
DAY_MICROSECOND
DAY_MINUTE
DAY_SECOND
DEC
DECIMAL
DECLARE
DEFAULT
DELAYED
DELETE
DENSE_RANK
DESC
DESCRIBE
DETERMINISTIC
DISTINCT
DISTINCTROW
DIV
DOUBLE
DROP
DUAL
EACH
ELSE
ELSEIF
EMPTY
ENCLOSED
ESCAPED
EXCEPT
EXISTS
EXIT
EXPLAIN
FALSE
FETCH
FIRST_VALUE
FLOAT
FLOAT4
FLOAT8
FOR
FORCE
FOREIGN
FROM
FULLTEXT
GENERATED
GET
GRANT
GROUP
GROUPING
GROUPS
HAVING
HIGH_PRIORITY
HOUR_MICROSECOND
HOUR_MINUTE
HOUR_SECOND
IF
IGNORE
IN
INDEX
INFILE
INNER
INOUT
INSENSITIVE
INSERT
INT
INT1
INT2
INT3
INT4
INT8
INTEGER
INTERSECT
INTERVAL
INTO
IO_AFTER_GTIDS
IO_BEFORE_GTIDS
IS
ITERATE
JOIN
JSON_TABLE
KEY
KEYS
KILL
LAG
LAST_VALUE
LATERAL
LEAD
LEADING
LEAVE
LEFT
LIKE
LIMIT
LINEAR
LINES
LOAD
LOCALTIME
LOCALTIMESTAMP
LOCK
LONG
LONGBLOB
LONGTEXT
LOOP
LOW_PRIORITY
MASTER_BIND
MASTER_SSL_VERIFY_SERVER_CERT
MATCH
MAXVALUE
MEDIUMBLOB
MEDIUMINT
MEDIUMTEXT
MIDDLEINT
MINUTE_MICROSECOND
MINUTE_SECOND
MOD
MODIFIES
NATURAL
NOT
NO_WRITE_TO_BINLOG
NTH_VALUE
NTILE
NULL
NUMERIC
OF
ON
OPTIMIZE
OPTIMIZER_COSTS
OPTION
OPTIONALLY
OR
ORDER
OUT
OUTER
OUTFILE
OVER
PARTITION
PERCENT_RANK
PRECISION
PRIMARY
PROCEDURE
PURGE
RANGE
RANK
READ
READS
READ_WRITE
REAL
RECURSIVE
REFERENCES
REGEXP
RELEASE
RENAME
REPEAT
REPLACE
REQUIRE
RESIGNAL
RESTRICT
RETURN
REVOKE
RIGHT
RLIKE
ROW_NUMBER
SCHEMA
SCHEMAS
SECOND_MICROSECOND
SELECT
SENSITIVE
SEPARATOR
SET
SHOW
SIGNAL
SMALLINT
SPATIAL
SPECIFIC
SQL
SQLEXCEPTION
SQLSTATE
SQLWARNING
SQL_BIG_RESULT
SQL_CALC_FOUND_ROWS
SQL_SMALL_RESULT
SSL
STARTING
STORED
STRAIGHT_JOIN
SYSTEM
TABLE
TERMINATED
THEN
TINYBLOB
TINYINT
TINYTEXT
TO
TRAILING
TRIGGER
TRUE
UNDO
UNION
UNIQUE
UNLOCK
UNSIGNED
UPDATE
USAGE
USE
USING
UTC_DATE
UTC_TIME
UTC_TIMESTAMP
VALUES
VARBINARY
VARCHAR
VARCHARACTER
VARYING
VIRTUAL
WHEN
WHERE
WHILE
WINDOW
WITH
WRITE
XOR
YEAR_MONTH
ZEROFILL
"""

mysql_unreserved_keywords = """ACCOUNT
ACTION
ACTIVE
ADMIN
AFTER
AGAINST
AGGREGATE
ALGORITHM
ALWAYS
ANALYSE
ANY
ARRAY
ASCII
AT
ATTRIBUTE
AUTHENTICATION
AUTOEXTEND_SIZE
AUTO_INCREMENT
AVG
AVG_ROW_LENGTH
BACKUP
BEGIN
BINLOG
BIT
BLOCK
BOOL
BOOLEAN
BTREE
BUCKETS
BYTE
CACHE
CASCADED
CATALOG_NAME
CHAIN
CHALLENGE_RESPONSE
CHANGED
CHANNEL
CHARSET
CHECKSUM
CIPHER
CLASS_ORIGIN
CLIENT
CLONE
CLOSE
COALESCE
CODE
COLLATION
COLUMNS
COLUMN_FORMAT
COLUMN_NAME
COMMENT
COMMIT
COMMITTED
COMPACT
COMPLETION
COMPONENT
COMPRESSED
COMPRESSION
CONCURRENT
CONNECTION
CONSISTENT
CONSTRAINT_CATALOG
CONSTRAINT_NAME
CONSTRAINT_SCHEMA
CONTAINS
CONTEXT
CPU
CUBE
CUME_DIST
CURRENT
CURSOR_NAME
DATA
DATAFILE
DATE
DATETIME
DAY
DEALLOCATE
DEFAULT_AUTH
DEFINER
DEFINITION
DELAY_KEY_WRITE
DENSE_RANK
DESCRIPTION
DES_KEY_FILE
DIAGNOSTICS
DIRECTORY
DISABLE
DISCARD
DISK
DO
DUMPFILE
DUPLICATE
DYNAMIC
EMPTY
ENABLE
ENCRYPTION
END
ENDS
ENFORCED
ENGINE
ENGINES
ENGINE_ATTRIBUTE
ENUM
ERROR
ERRORS
ESCAPE
EVENT
EVENTS
EVERY
EXCHANGE
EXCLUDE
EXECUTE
EXPANSION
EXPIRE
EXPORT
EXTENDED
EXTENT_SIZE
FACTOR
FAILED_LOGIN_ATTEMPTS
FAST
FAULTS
FIELDS
FILE
FILE_BLOCK_SIZE
FILTER
FINISH
FIRST
FIRST_VALUE
FIXED
FLUSH
FOLLOWING
FOLLOWS
FORMAT
FOUND
FULL
FUNCTION
GENERAL
GEOMCOLLECTION
GEOMETRY
GEOMETRYCOLLECTION
GET_FORMAT
GET_MASTER_PUBLIC_KEY
GET_SOURCE_PUBLIC_KEY
GLOBAL
GRANTS
GROUPING
GROUPS
GROUP_REPLICATION
GTID_ONLY
HANDLER
HASH
HELP
HISTOGRAM
HISTORY
HOST
HOSTS
HOUR
IDENTIFIED
IGNORE_SERVER_IDS
IMPORT
INACTIVE
INDEXES
INITIAL
INITIAL_SIZE
INITIATE
INSERT_METHOD
INSTALL
INSTANCE
INTERSECT
INVISIBLE
INVOKER
IO
IO_THREAD
IPC
ISOLATION
ISSUER
JSON
JSON_TABLE
JSON_VALUE
KEYRING
KEY_BLOCK_SIZE
LAG
LANGUAGE
LAST
LAST_VALUE
LATERAL
LEAD
LEAVES
LESS
LEVEL
LINESTRING
LIST
LOCAL
LOCKED
LOCKS
LOGFILE
LOGS
MASTER
MASTER_AUTO_POSITION
MASTER_COMPRESSION_ALGORITHMS
MASTER_CONNECT_RETRY
MASTER_DELAY
MASTER_HEARTBEAT_PERIOD
MASTER_HOST
MASTER_LOG_FILE
MASTER_LOG_POS
MASTER_PASSWORD
MASTER_PORT
MASTER_PUBLIC_KEY_PATH
MASTER_RETRY_COUNT
MASTER_SERVER_ID
MASTER_SSL
MASTER_SSL_CA
MASTER_SSL_CAPATH
MASTER_SSL_CERT
MASTER_SSL_CIPHER
MASTER_SSL_CRL
MASTER_SSL_CRLPATH
MASTER_SSL_KEY
MASTER_TLS_CIPHERSUITES
MASTER_TLS_VERSION
MASTER_USER
MASTER_ZSTD_COMPRESSION_LEVEL
MAX_CONNECTIONS_PER_HOUR
MAX_QUERIES_PER_HOUR
MAX_ROWS
MAX_SIZE
MAX_UPDATES_PER_HOUR
MAX_USER_CONNECTIONS
MEDIUM
MEMBER
MEMORY
MERGE
MESSAGE_TEXT
MICROSECOND
MIGRATE
MINUTE
MIN_ROWS
MODE
MODIFY
MONTH
MULTILINESTRING
MULTIPOINT
MULTIPOLYGON
MUTEX
MYSQL_ERRNO
NAME
NAMES
NATIONAL
NCHAR
NDB
NDBCLUSTER
NESTED
NETWORK_NAMESPACE
NEVER
NEW
NEXT
NO
NODEGROUP
NONE
NOWAIT
NO_WAIT
NTH_VALUE
NTILE
NULLS
NUMBER
NVARCHAR
OF
OFF
OFFSET
OJ
OLD
ONE
ONLY
OPEN
OPTIONAL
OPTIONS
ORDINALITY
ORGANIZATION
OTHERS
OVER
OWNER
PACK_KEYS
PAGE
PARSER
PARSE_GCOL_EXPR
PARTIAL
PARTITIONING
PARTITIONS
PASSWORD
PASSWORD_LOCK_TIME
PATH
PERCENT_RANK
PERSIST
PERSIST_ONLY
PHASE
PLUGIN
PLUGINS
PLUGIN_DIR
POINT
POLYGON
PORT
PRECEDES
PRECEDING
PREPARE
PRESERVE
PREV
PRIVILEGES
PRIVILEGE_CHECKS_USER
PROCESS
PROCESSLIST
PROFILE
PROFILES
PROXY
QUARTER
QUERY
QUICK
RANDOM
RANK
READ_ONLY
REBUILD
RECOVER
RECURSIVE
REDOFILE
REDO_BUFFER_SIZE
REDUNDANT
REFERENCE
REGISTRATION
RELAY
RELAYLOG
RELAY_LOG_FILE
RELAY_LOG_POS
RELAY_THREAD
RELOAD
REMOTE
REMOVE
REORGANIZE
REPAIR
REPEATABLE
REPLICA
REPLICAS
REPLICATE_DO_DB
REPLICATE_DO_TABLE
REPLICATE_IGNORE_DB
REPLICATE_IGNORE_TABLE
REPLICATE_REWRITE_DB
REPLICATE_WILD_DO_TABLE
REPLICATE_WILD_IGNORE_TABLE
REPLICATION
REQUIRE_ROW_FORMAT
RESET
RESOURCE
RESPECT
RESTART
RESTORE
RESUME
RETAIN
RETURNED_SQLSTATE
RETURNING
RETURNS
REUSE
REVERSE
ROLE
ROLLBACK
ROLLUP
ROTATE
ROUTINE
ROW
ROWS
ROW_COUNT
ROW_FORMAT
ROW_NUMBER
RTREE
SAVEPOINT
SCHEDULE
SCHEMA_NAME
SECOND
SECONDARY
SECONDARY_ENGINE
SECONDARY_ENGINE_ATTRIBUTE
SECONDARY_LOAD
SECONDARY_UNLOAD
SECURITY
SERIAL
SERIALIZABLE
SERVER
SESSION
SHARE
SHUTDOWN
SIGNED
SIMPLE
SKIP
SLAVE
SLOW
SNAPSHOT
SOCKET
SOME
SONAME
SOUNDS
SOURCE
SOURCE_AUTO_POSITION
SOURCE_BIND
SOURCE_COMPRESSION_ALGORITHMS
SOURCE_CONNECT_RETRY
SOURCE_DELAY
SOURCE_HEARTBEAT_PERIOD
SOURCE_HOST
SOURCE_LOG_FILE
SOURCE_LOG_POS
SOURCE_PASSWORD
SOURCE_PORT
SOURCE_PUBLIC_KEY_PATH
SOURCE_RETRY_COUNT
SOURCE_SSL
SOURCE_SSL_CA
SOURCE_SSL_CAPATH
SOURCE_SSL_CERT
SOURCE_SSL_CIPHER
SOURCE_SSL_CRL
SOURCE_SSL_CRLPATH
SOURCE_SSL_KEY
SOURCE_SSL_VERIFY_SERVER_CERT
SOURCE_TLS_CIPHERSUITES
SOURCE_TLS_VERSION
SOURCE_USER
SOURCE_ZSTD_COMPRESSION_LEVEL
SQL_AFTER_GTIDS
SQL_AFTER_MTS_GAPS
SQL_BEFORE_GTIDS
SQL_BUFFER_RESULT
SQL_CACHE
SQL_NO_CACHE
SQL_THREAD
SQL_TSI_DAY
SQL_TSI_HOUR
SQL_TSI_MINUTE
SQL_TSI_MONTH
SQL_TSI_QUARTER
SQL_TSI_SECOND
SQL_TSI_WEEK
SQL_TSI_YEAR
SRID
STACKED
START
STARTS
STATS_AUTO_RECALC
STATS_PERSISTENT
STATS_SAMPLE_PAGES
STATUS
STOP
STORAGE
STREAM
STRING
SUBCLASS_ORIGIN
SUBJECT
SUBPARTITION
SUBPARTITIONS
SUPER
SUSPEND
SWAPS
SWITCHES
SYSTEM
TABLES
TABLESPACE
TABLE_CHECKSUM
TABLE_NAME
TEMPORARY
TEMPTABLE
TEXT
THAN
THREAD_PRIORITY
TIES
TIME
TIMESTAMP
TIMESTAMPADD
TIMESTAMPDIFF
TLS
TRANSACTION
TRIGGERS
TRUNCATE
TYPE
TYPES
UNBOUNDED
UNCOMMITTED
UNDEFINED
UNDOFILE
UNDO_BUFFER_SIZE
UNICODE
UNINSTALL
UNKNOWN
UNREGISTER
UNTIL
UPGRADE
USER
USER_RESOURCES
USE_FRM
VALIDATION
VALUE
VARIABLES
VCPU
VIEW
VISIBLE
WAIT
WARNINGS
WEEK
WEIGHT_STRING
WINDOW
WITHOUT
WORK
WRAPPER
X509
XA
XID
XML
YEAR
ZONE
"""

# These are not MySQL keywords, but SQLFluff needs them to parse well.
mysql_unreserved_keywords += """NOW
SHARED
INPLACE
"""



"""Tests issue #1373 doesn't reoccur.

The combination of L003 (incorrect indentation), L036 (select targets),
and L039 (unnecessary white space) can result in incorrect indentation.
"""

import sqlfluff


def test__rules__std_L003_L036_L039():
    """Verify that double indents don't flag L039."""
    sql = """
    WITH example AS (
        SELECT my_id,
            other_thing,
            one_more
        FROM
            my_table
    )

    SELECT my_id
    FROM example\n"""
    fixed_sql = """
    WITH example AS (
        SELECT
            my_id,
            other_thing,
            one_more
        FROM
            my_table
    )

    SELECT my_id
    FROM example\n"""
    result = sqlfluff.fix(sql, exclude_rules=["L050"])
    assert result == fixed_sql



"""Static methods to support ReflowSequence.rebreak()."""

import logging
from dataclasses import dataclass

from typing import List, Type, Tuple, cast

from sqlfluff.core.parser import BaseSegment
from sqlfluff.core.rules import LintFix, LintResult

from sqlfluff.utils.reflow.elements import ReflowBlock, ReflowPoint, ReflowSequenceType
from sqlfluff.utils.reflow.reindent import deduce_line_indent
from sqlfluff.utils.reflow.helpers import fixes_from_results, pretty_segment_name


# We're in the utils module, but users will expect reflow
# logs to appear in the context of rules. Hence it's a subset
# of the rules logger.
reflow_logger = logging.getLogger("sqlfluff.rules.reflow")


@dataclass(frozen=True)
class _RebreakSpan:
    """A location within a sequence to consider rebreaking."""

    target: BaseSegment
    start_idx: int
    end_idx: int
    line_position: str
    strict: bool


@dataclass(frozen=True)
class _RebreakIndices:
    """Indices of points for a _RebreakLocation."""

    dir: int
    adj_pt_idx: int
    newline_pt_idx: int
    pre_code_pt_idx: int

    @classmethod
    def from_elements(
        cls: Type["_RebreakIndices"],
        elements: ReflowSequenceType,
        start_idx: int,
        dir: int,
    ) -> "_RebreakIndices":
        """Iterate through the elements to deduce important point indices."""
        assert dir in (1, -1), "Direction must be a unit direction (i.e. 1 or -1)."
        # Limit depends on the direction
        limit = 0 if dir == -1 else len(elements)
        # The adjacent point is just the next one.
        adj_point_idx = start_idx + dir
        # The newline point is next. We hop in 2s because we're checking
        # only points, which alternate with blocks.
        for newline_point_idx in range(adj_point_idx, limit, 2 * dir):
            if "newline" in elements[newline_point_idx].class_types or any(
                seg.is_code for seg in elements[newline_point_idx + dir].segments
            ):
                break
        # Finally we look for the point preceding the next code element.
        for pre_code_point_idx in range(newline_point_idx, limit, 2 * dir):
            if any(seg.is_code for seg in elements[pre_code_point_idx + dir].segments):
                break
        return cls(dir, adj_point_idx, newline_point_idx, pre_code_point_idx)


@dataclass(frozen=True)
class _RebreakLocation:
    """A location within a sequence to rebreak, with metadata."""

    target: BaseSegment
    prev: _RebreakIndices
    next: _RebreakIndices
    line_position: str
    strict: bool

    @classmethod
    def from_span(
        cls: Type["_RebreakLocation"], span: _RebreakSpan, elements: ReflowSequenceType
    ) -> "_RebreakLocation":
        """Expand a span to a location."""
        return cls(
            span.target,
            _RebreakIndices.from_elements(elements, span.start_idx, -1),
            _RebreakIndices.from_elements(elements, span.end_idx, 1),
            span.line_position,
            span.strict,
        )

    def pretty_target_name(self) -> str:
        """Get a nicely formatted name of the target."""
        return pretty_segment_name(self.target)

    def has_templated_newline(self, elements: ReflowSequenceType) -> bool:
        """Is either side a templated newline?

        If either side has a templated newline, then that's ok too.
        The intent here is that if the next newline is a _templated_
        one, then in the source there will be a tag ({{ tag }}), which
        acts like _not having a newline_.
        """
        # Check the _last_ newline of the previous point.
        # Slice backward to search in reverse.
        for seg in elements[self.prev.newline_pt_idx].segments[::-1]:
            if seg.is_type("newline"):
                if not seg.pos_marker.is_literal():
                    return True
                break
        # Check the _first_ newline of the next point.
        for seg in elements[self.next.newline_pt_idx].segments:
            if seg.is_type("newline"):
                if not seg.pos_marker.is_literal():
                    return True
                break
        return False

    def has_inappropriate_newlines(
        self, elements: ReflowSequenceType, strict: bool = False
    ) -> bool:
        """Is the span surrounded by one (but not two) line breaks?

        Args:
            elements: The elements of the ReflowSequence this element
               is taken from to allow comparison.
            strict (:obj:`bool`): If set to true, this will not allow
               the case where there aren't newlines on either side.
        """
        # Here we use the newline index, not
        # just the adjacent point, so that we can see past comments.
        n_prev_newlines = elements[self.prev.newline_pt_idx].num_newlines()
        n_next_newlines = elements[self.next.newline_pt_idx].num_newlines()
        newlines_on_neither_side = n_prev_newlines + n_next_newlines == 0
        newlines_on_both_sides = n_prev_newlines > 0 and n_next_newlines > 0
        return (
            # If there isn't a newline on either side then carry
            # on, unless it's strict.
            (newlines_on_neither_side and not strict)
            # If there is a newline on BOTH sides. That's ok.
            or newlines_on_both_sides
        )


def identify_rebreak_spans(
    element_buffer: ReflowSequenceType, root_segment: BaseSegment
) -> List[_RebreakSpan]:
    """Identify areas in file to rebreak.

    A span here is a block, or group of blocks which have
    explicit configs for their line position, either directly
    as raw segments themselves or by virtue of one of their
    parent segments.
    """
    spans: List[_RebreakSpan] = []
    # We'll need at least two elements each side, so constrain
    # our range accordingly.
    for idx in range(2, len(element_buffer) - 2):
        # Only evaluate blocks:
        elem = element_buffer[idx]
        # Only evaluate blocks
        if not isinstance(elem, ReflowBlock):
            continue
        # Does the element itself have config? (The easy case)
        if elem.line_position:
            # Blocks should only have one segment so it's easy to pick it.
            spans.append(
                _RebreakSpan(
                    elem.segments[0],
                    idx,
                    idx,
                    # NOTE: this isn't pretty but until it needs to be more
                    # complex, this works.
                    elem.line_position.split(":")[0],
                    elem.line_position.endswith("strict"),
                )
            )
        # Do any of its parents have config, and are we at the start
        # of them?
        for key in elem.line_position_configs.keys():
            # If we're not at the start of the segment, then pass.
            if elem.depth_info.stack_positions[key].idx != 0:
                continue
            # Can we find the end?
            for end_idx in range(idx, len(element_buffer) - 2):
                end_elem = element_buffer[end_idx]
                if not isinstance(end_elem, ReflowBlock):
                    continue
                if end_elem.depth_info.stack_positions[key].type in ("end", "solo"):
                    # Found the end. Add it to the stack.
                    # We reference the appropriate element from the parent stack.
                    target_depth = elem.depth_info.stack_hashes.index(key)
                    target = root_segment.path_to(element_buffer[idx].segments[0])[
                        target_depth
                    ].segment
                    spans.append(
                        _RebreakSpan(
                            target,
                            idx,
                            end_idx,
                            # NOTE: this isn't pretty but until it needs to be more
                            # complex, this works.
                            elem.line_position_configs[key].split(":")[0],
                            elem.line_position_configs[key].endswith("strict"),
                        )
                    )
                    break
            # If we find the start, but not the end, it's not a problem, but
            # we won't be rebreaking this span. This is important so that we
            # don't rebreak part of something without the context of what's
            # in the rest of it. We continue without adding it to the buffer.
    return spans


def rebreak_sequence(
    elements: ReflowSequenceType, root_segment: BaseSegment
) -> Tuple[ReflowSequenceType, List[LintResult]]:
    """Reflow line breaks within a sequence.

    Initially this only _moves_ existing segments
    around line breaks (e.g. for operators and commas),
    but eventually this method should also handle line
    length considerations too.

    This intentionally does *not* handle indentation,
    as the existing indents are assumed to be correct.
    """
    lint_results: List[LintResult] = []
    fixes: List[LintFix] = []
    elem_buff: ReflowSequenceType = elements.copy()

    # Given a sequence we should identify the objects which
    # make sense to rebreak. That includes any raws with config,
    # but also and parent segments which have config and we can
    # find both ends for. Given those spans, we then need to find
    # the points either side of them and then the blocks either
    # side to respace them at the same time.

    # 1. First find appropriate spans.
    spans = identify_rebreak_spans(elem_buff, root_segment)

    # The spans give us the edges of operators, but for line positioning we need
    # to handle comments differently. There are two other important points:
    # 1. The next newline outward before code (but passing over comments).
    # 2. The point before the next _code_ segment (ditto comments).
    locations = [_RebreakLocation.from_span(span, elem_buff) for span in spans]

    # Handle each span:
    for loc in locations:
        reflow_logger.debug(
            "Handing Rebreak Span (%r: %s): %r",
            loc.line_position,
            loc.target,
            "".join(
                elem.raw
                for elem in elem_buff[
                    loc.prev.pre_code_pt_idx - 1 : loc.next.pre_code_pt_idx + 2
                ]
            ),
        )

        if loc.has_inappropriate_newlines(elem_buff, strict=loc.strict):
            continue

        if loc.has_templated_newline(elem_buff):
            continue

        # Points and blocks either side are just offsets from the indices.
        prev_point = cast(ReflowPoint, elem_buff[loc.prev.adj_pt_idx])
        next_point = cast(ReflowPoint, elem_buff[loc.next.adj_pt_idx])

        # So we know we have a preference, is it ok?
        if loc.line_position == "leading":
            if elem_buff[loc.prev.newline_pt_idx].num_newlines():
                # We're good. It's already leading.
                continue

            # Generate the text for any issues.
            pretty_name = loc.pretty_target_name()
            if loc.strict:  # pragma: no cover
                # TODO: The 'strict' option isn't widely tested yet.
                desc = f"{pretty_name.capitalize()} should always start a new line."
            else:
                desc = (
                    f"Found trailing {pretty_name}. Expected only leading "
                    "near line breaks."
                )

            # Is it the simple case with no comments between the
            # old and new desired locations and only a single following
            # whitespace?
            if (
                loc.next.adj_pt_idx == loc.next.pre_code_pt_idx
                and elem_buff[loc.next.newline_pt_idx].num_newlines() == 1
            ):
                reflow_logger.debug("  Trailing Easy Case")
                # Simple case. No comments.
                # Strip newlines from the next point. Apply the indent to
                # the previous point.
                new_results, prev_point = prev_point.indent_to(
                    next_point.get_indent() or "", before=loc.target
                )
                new_results, next_point = next_point.respace_point(
                    cast(ReflowBlock, elem_buff[loc.next.adj_pt_idx - 1]),
                    cast(ReflowBlock, elem_buff[loc.next.adj_pt_idx + 1]),
                    root_segment=root_segment,
                    lint_results=new_results,
                    strip_newlines=True,
                )

                # Update the points in the buffer
                elem_buff[loc.prev.adj_pt_idx] = prev_point
                elem_buff[loc.next.adj_pt_idx] = next_point
            else:
                reflow_logger.debug("  Trailing Tricky Case")
                # Otherwise we've got a tricky scenario where there are comments
                # to negotiate around. In this case, we _move the target_
                # rather than just adjusting the whitespace.

                # Delete the existing position of the target, and
                # the _preceding_ point.
                fixes.append(LintFix.delete(loc.target))
                for seg in elem_buff[loc.prev.adj_pt_idx].segments:
                    fixes.append(LintFix.delete(seg))

                # We always reinsert after the first point, but respace
                # the inserted point to ensure it's the right size given
                # configs.
                new_results, new_point = ReflowPoint(()).respace_point(
                    cast(ReflowBlock, elem_buff[loc.next.adj_pt_idx - 1]),
                    cast(ReflowBlock, elem_buff[loc.next.pre_code_pt_idx + 1]),
                    root_segment=root_segment,
                    lint_results=[],
                    anchor_on="after",
                )
                fixes.append(
                    LintFix.create_after(
                        elem_buff[loc.next.pre_code_pt_idx].segments[-1],
                        [loc.target],
                    )
                )

                elem_buff = (
                    elem_buff[: loc.prev.adj_pt_idx]
                    + elem_buff[loc.next.adj_pt_idx : loc.next.pre_code_pt_idx + 1]
                    + elem_buff[
                        loc.prev.adj_pt_idx + 1 : loc.next.adj_pt_idx
                    ]  # the target
                    + [new_point]
                    + elem_buff[loc.next.pre_code_pt_idx + 1 :]
                )

        elif loc.line_position == "trailing":
            if elem_buff[loc.next.newline_pt_idx].num_newlines():
                # We're good, it's already trailing.
                continue

            # Generate the text for any issues.
            pretty_name = loc.pretty_target_name()
            if loc.strict:  # pragma: no cover
                # TODO: The 'strict' option isn't widely tested yet.
                desc = (
                    f"{pretty_name.capitalize()} should always be at the end of a line."
                )
            else:
                desc = (
                    f"Found leading {pretty_name}. Expected only trailing "
                    "near line breaks."
                )

            # Is it the simple case with no comments between the
            # old and new desired locations and only one previous newline?
            if (
                loc.prev.adj_pt_idx == loc.prev.pre_code_pt_idx
                and elem_buff[loc.prev.newline_pt_idx].num_newlines() == 1
            ):
                reflow_logger.debug("  Leading Easy Case")
                # Simple case. No comments.
                # Strip newlines from the previous point. Apply the indent
                # to the next point.
                new_results, next_point = next_point.indent_to(
                    prev_point.get_indent() or "", after=loc.target
                )
                new_results, prev_point = prev_point.respace_point(
                    cast(ReflowBlock, elem_buff[loc.prev.adj_pt_idx - 1]),
                    cast(ReflowBlock, elem_buff[loc.prev.adj_pt_idx + 1]),
                    root_segment=root_segment,
                    lint_results=new_results,
                    strip_newlines=True,
                )

                # Update the points in the buffer
                elem_buff[loc.prev.adj_pt_idx] = prev_point
                elem_buff[loc.next.adj_pt_idx] = next_point
            else:
                reflow_logger.debug("  Leading Tricky Case")
                # Otherwise we've got a tricky scenario where there are comments
                # to negotiate around. In this case, we _move the target_
                # rather than just adjusting the whitespace.

                # Delete the existing position of the target, and
                # the _following_ point.
                fixes.append(LintFix.delete(loc.target))
                for seg in elem_buff[loc.next.adj_pt_idx].segments:
                    fixes.append(LintFix.delete(seg))

                # We always reinsert before the first point, but respace
                # the inserted point to ensure it's the right size given
                # configs.
                new_results, new_point = ReflowPoint(()).respace_point(
                    cast(ReflowBlock, elem_buff[loc.prev.pre_code_pt_idx - 1]),
                    cast(ReflowBlock, elem_buff[loc.prev.adj_pt_idx + 1]),
                    root_segment=root_segment,
                    lint_results=[],
                    anchor_on="before",
                )
                fixes.append(
                    LintFix.create_before(
                        elem_buff[loc.prev.pre_code_pt_idx].segments[0],
                        [loc.target],
                    )
                )

                elem_buff = (
                    elem_buff[: loc.prev.pre_code_pt_idx]
                    + [new_point]
                    + elem_buff[
                        loc.prev.adj_pt_idx + 1 : loc.next.adj_pt_idx
                    ]  # the target
                    + elem_buff[loc.prev.pre_code_pt_idx : loc.prev.adj_pt_idx + 1]
                    + elem_buff[loc.next.adj_pt_idx + 1 :]
                )

        elif loc.line_position == "alone":
            # If we get here we can assume that the element is currently
            # either leading or trailing and needs to be moved onto its
            # own line.

            # Generate the text for any issues.
            pretty_name = loc.pretty_target_name()
            desc = (
                f"{pretty_name.capitalize()}s should always have a line break "
                "both before and after."
            )

            # First handle the following newlines first (easy).
            if not elem_buff[loc.next.newline_pt_idx].num_newlines():
                reflow_logger.debug("  Found missing newline after in alone case")
                new_results, next_point = next_point.indent_to(
                    deduce_line_indent(loc.target.raw_segments[-1], root_segment),
                    after=loc.target,
                )
                # Update the point in the buffer
                elem_buff[loc.next.adj_pt_idx] = next_point

            # Then handle newlines before. (hoisting past comments if needed).
            if not elem_buff[loc.prev.adj_pt_idx].num_newlines():
                reflow_logger.debug("  Found missing newline before in alone case")
                # NOTE: In the case that there are comments _after_ the
                # target, they will be moved with it. This might break things
                # but there isn't an unambiguous way to do this, because we
                # can't be sure what the comments are referring to.
                # Given that, we take the simple option.
                new_results, prev_point = prev_point.indent_to(
                    deduce_line_indent(loc.target.raw_segments[0], root_segment),
                    before=loc.target,
                )
                # Update the point in the buffer
                elem_buff[loc.prev.adj_pt_idx] = prev_point

        else:
            raise NotImplementedError(  # pragma: no cover
                f"Unexpected line_position config: {loc.line_position}"
            )

        # Consolidate results and consume fix buffer
        lint_results.append(
            LintResult(
                loc.target,
                fixes=fixes_from_results(new_results) + fixes,
                description=desc,
            )
        )
        fixes = []

    return elem_buff, lint_results



"""Dataclasses for reflow work."""

from itertools import chain
import logging
from dataclasses import dataclass
from typing import Dict, List, Optional, Sequence, Set, Tuple, Type, Union, cast

from sqlfluff.core.parser import BaseSegment, RawSegment
from sqlfluff.core.parser.segments import (
    NewlineSegment,
    WhitespaceSegment,
    TemplateSegment,
)
from sqlfluff.core.rules.base import LintFix, LintResult

from sqlfluff.utils.reflow.config import ReflowConfig
from sqlfluff.utils.reflow.depthmap import DepthInfo

# Respace Algorithms
from sqlfluff.utils.reflow.respace import (
    determine_constraints,
    process_spacing,
    handle_respace__inline_with_space,
    handle_respace__inline_without_space,
)

# We're in the utils module, but users will expect reflow
# logs to appear in the context of rules. Hence it's a subset
# of the rules logger.
reflow_logger = logging.getLogger("sqlfluff.rules.reflow")


@dataclass(frozen=True)
class ReflowElement:
    """Base reflow element class."""

    segments: Tuple[RawSegment, ...]

    @staticmethod
    def _class_types(segments: Sequence[RawSegment]) -> Set[str]:
        return set(chain.from_iterable(seg.class_types for seg in segments))

    @property
    def class_types(self) -> Set[str]:
        """Get the set of contained class types.

        Parallel to `BaseSegment.class_types`
        """
        return self._class_types(self.segments)

    @property
    def raw(self) -> str:
        """Get the current raw representation."""
        return "".join(seg.raw for seg in self.segments)

    def num_newlines(self) -> int:
        """Return the number of newlines in this element."""
        return sum(bool("newline" in seg.class_types) for seg in self.segments)


@dataclass(frozen=True)
class ReflowBlock(ReflowElement):
    """Class for keeping track of elements to reflow.

    This class, and its sibling :obj:`ReflowPoint`, should not
    normally be manipulated directly by rules, but instead should
    be manipulated using :obj:`ReflowSequence`.

    It holds segments to reflow and also exposes configuration
    regarding how they are expected to reflow around others. Typically
    it holds only a single element, which is usually code or a
    templated element. Because reflow operations control spacing,
    it would be very unusual for this object to be modified; as
    such it exposes relatively few methods.

    The attributes exposed are designed to be "post configuration"
    i.e. they should reflect configuration appropriately.
    """

    #: Desired spacing before this block.
    #: See :ref:`layoutspacingconfig`
    spacing_before: str
    #: Desired spacing after this block.
    #: See :ref:`layoutspacingconfig`
    spacing_after: str
    #: Desired line position for this block.
    #: See :ref:`layoutspacingconfig`
    line_position: Optional[str]
    #: Metadata on the depth of this segment within the parse tree
    #: which is used in inferring how and where line breaks should
    #: exist.
    depth_info: DepthInfo
    #: Desired spacing configurations for parent segments
    #: of the segment in this block.
    #: See :ref:`layoutspacingconfig`
    stack_spacing_configs: Dict[int, str]
    #: Desired line position configurations for parent segments
    #: of the segment in this block.
    #: See :ref:`layoutspacingconfig`
    line_position_configs: Dict[int, str]

    @classmethod
    def from_config(
        cls: Type["ReflowBlock"], segments, config: ReflowConfig, depth_info: DepthInfo
    ) -> "ReflowBlock":
        """Construct a ReflowBlock while extracting relevant configuration.

        This is the primary route to construct a ReflowBlock, as
        is allows all of the inference of the spacing and position
        configuration from the segments it contains and the
        appropriate config objects.
        """
        block_config = config.get_block_config(cls._class_types(segments), depth_info)
        stack_spacing_configs = {}
        line_position_configs = {}
        for hash, class_types in zip(
            depth_info.stack_hashes, depth_info.stack_class_types
        ):
            cfg = config.get_block_config(class_types)
            if cfg.spacing_within:
                stack_spacing_configs[hash] = cfg.spacing_within
            if cfg.line_position:
                line_position_configs[hash] = cfg.line_position
        return cls(
            segments=segments,
            spacing_before=block_config.spacing_before,
            spacing_after=block_config.spacing_after,
            line_position=block_config.line_position,
            depth_info=depth_info,
            stack_spacing_configs=stack_spacing_configs,
            line_position_configs=line_position_configs,
        )


def _indent_description(indent: str):
    """Construct a human readable description of the indent."""
    if indent == "":
        return "no indent"
    elif indent[0] == " ":
        assert all(c == " " for c in indent)
        return f"indent of {len(indent)} spaces"
    elif indent[0] == "\t":  # pragma: no cover
        assert all(c == "\t" for c in indent)
        return f"indent of {len(indent)} tabs"
    else:  # pragma: no cover
        raise NotImplementedError(f"Invalid indent construction: {indent!r}")


@dataclass(frozen=True)
class ReflowPoint(ReflowElement):
    """Class for keeping track of editable elements in reflow.

    This class, and its sibling :obj:`ReflowBlock`, should not
    normally be manipulated directly by rules, but instead should
    be manipulated using :obj:`ReflowSequence`.

    It holds segments which can be changed during a reflow operation
    such as whitespace and newlines.It may also contain :obj:`Indent`
    and :obj:`Dedent` elements.

    It holds no configuration and is influenced by the blocks on either
    side, so that any operations on it usually have that configuration
    passed in as required.
    """

    def _get_indent_segment(self) -> Optional[RawSegment]:
        """Get the current indent segment (if there)."""
        indent = None
        for seg in reversed(self.segments):
            if seg.is_type("newline"):
                return indent
            elif seg.is_type("whitespace"):
                indent = seg
        # i.e. if we never find a newline, it's not an indent.
        return None

    def get_indent(self) -> Optional[str]:
        """Get the current indent (if there)."""
        seg = self._get_indent_segment()
        return seg.raw if seg else None

    def indent_to(
        self,
        desired_indent: str,
        after: Optional[BaseSegment] = None,
        before: Optional[BaseSegment] = None,
    ) -> Tuple[List[LintResult], "ReflowPoint"]:
        """Coerce a point to have a particular indent.

        If the point currently contains no newlines, one will
        be introduced and any trailing whitespace will be effectively
        removed.

        More specifically, the newline is *inserted before* the existing
        whitespace, with the new indent being a *replacement* for that
        same whitespace.
        """
        # Get the indent (or in the case of no newline, the last whitespace)
        indent_seg = self._get_indent_segment()
        reflow_logger.debug(
            "Coercing indent %s to %r. (newlines: %s)",
            indent_seg,
            desired_indent,
            self.num_newlines(),
        )
        if self.num_newlines():
            # There is already a newline.
            if indent_seg:
                # Coerce existing indent to desired.
                if indent_seg.raw == desired_indent:
                    # Trivial case. Indent already correct
                    return [], self
                elif desired_indent == "":
                    # Coerce to no indent. We don't want the indent. Delete it.
                    new_indent = indent_seg.edit(desired_indent)
                    idx = self.segments.index(indent_seg)
                    return [
                        LintResult(
                            indent_seg,
                            [LintFix.delete(indent_seg)],
                            description="Line should not be indented.",
                        )
                    ], ReflowPoint(self.segments[:idx] + self.segments[idx + 1 :])

                # Standard case of an indent change.
                new_indent = indent_seg.edit(desired_indent)
                idx = self.segments.index(indent_seg)
                return [
                    LintResult(
                        indent_seg,
                        [LintFix.replace(indent_seg, [new_indent])],
                        description=f"Expected {_indent_description(desired_indent)}.",
                    )
                ], ReflowPoint(
                    self.segments[:idx] + (new_indent,) + self.segments[idx + 1 :]
                )

            else:
                # There is a newline, but no indent. Make one after the newline
                # Find the index of the last newline.
                for idx in range(len(self.segments) - 1, 0, -1):
                    if self.segments[idx].is_type("newline"):
                        break
                new_indent = WhitespaceSegment(desired_indent)
                return [
                    LintResult(
                        self.segments[idx],
                        [LintFix.create_after(self.segments[idx], [new_indent])],
                        description=f"Expected {_indent_description(desired_indent)}.",
                    )
                ], ReflowPoint(
                    self.segments[: idx + 1] + (new_indent,) + self.segments[idx + 1 :]
                )

        else:
            # There isn't currently a newline.
            new_newline = NewlineSegment()
            # Check for whitespace
            ws_seg = None
            for seg in self.segments[::-1]:
                if seg.is_type("whitespace"):
                    ws_seg = seg
            if not ws_seg:
                # There isn't a whitespace segment either. We need to insert one.
                # Do we have an anchor?
                new_indent = WhitespaceSegment(desired_indent)
                if not before and not after:  # pragma: no cover
                    raise NotImplementedError(
                        "Not set up to handle empty points in this "
                        "scenario without provided before/after "
                        f"anchor: {self.segments}"
                    )
                # Otherwise make a new indent, attached to the relevant anchor.
                # Prefer anchoring before because it makes the labelling better.
                elif before:
                    before_raw = (
                        cast(TemplateSegment, before).source_str
                        if before.is_type("placeholder")
                        else before.raw
                    )
                    fix = LintFix.create_before(
                        before,
                        [new_newline, new_indent],
                    )
                    description = (
                        "Expected line break and "
                        f"{_indent_description(desired_indent)} "
                        f"before {before_raw!r}."
                    )
                else:
                    assert after  # mypy hint
                    after_raw = (
                        cast(TemplateSegment, after).source_str
                        if after.is_type("placeholder")
                        else after.raw
                    )
                    fix = LintFix.create_after(
                        after,
                        [new_newline, new_indent],
                    )
                    description = (
                        "Expected line break and "
                        f"{_indent_description(desired_indent)} "
                        f"after {after_raw!r}."
                    )
                new_point = ReflowPoint((new_newline, new_indent))
                anchor = before
            else:
                # There is whitespace. Coerce it to the right indent and add
                # a newline _before_. In the edge case that we're coercing to
                # _no indent_, edit existing indent to be the newline and leave
                # it there.
                new_segs: List[RawSegment]
                if desired_indent == "":
                    new_segs = [new_newline]
                else:
                    new_segs = [new_newline, ws_seg.edit(desired_indent)]
                idx = self.segments.index(ws_seg)
                # Prefer before, because it makes the anchoring better.
                if before:
                    description = (
                        "Expected line break and "
                        f"{_indent_description(desired_indent)} "
                        f"before {before.raw!r}."
                    )
                elif after:
                    description = (
                        "Expected line break and "
                        f"{_indent_description(desired_indent)} "
                        f"after {after.raw!r}."
                    )
                else:  # pragma: no cover
                    # NOTE: Doesn't have test coverage because there's
                    # normally an `after` or `before` value, so this
                    # clause is unused.
                    description = (
                        "Expected line break and "
                        f"{_indent_description(desired_indent)}."
                    )
                fix = LintFix.replace(ws_seg, new_segs)
                new_point = ReflowPoint(
                    self.segments[:idx] + tuple(new_segs) + self.segments[idx + 1 :]
                )
                anchor = ws_seg

            return [LintResult(anchor, fixes=[fix], description=description)], new_point

    def respace_point(
        self,
        prev_block: Optional[ReflowBlock],
        next_block: Optional[ReflowBlock],
        root_segment: BaseSegment,
        lint_results: List[LintResult],
        strip_newlines: bool = False,
        anchor_on: str = "before",
    ) -> Tuple[List[LintResult], "ReflowPoint"]:
        """Respace a point based on given constraints.

        NB: This effectively includes trailing whitespace fixes.

        Deletion and edit fixes are generated immediately, but creations
        are paused to the end and done in bulk so as not to generate conflicts.

        Note that the `strip_newlines` functionality exists here as a slight
        exception to pure respacing, but as a very simple case of positioning
        line breaks. The default operation of `respace` does not enable it,
        however it exists as a convenience for rules which wish to use it.
        """
        existing_results = lint_results[:]
        pre_constraint, post_constraint, strip_newlines = determine_constraints(
            prev_block, next_block, strip_newlines
        )

        reflow_logger.debug("Respacing: %s", self)

        # The buffer is used to create the new reflow point to return
        segment_buffer, last_whitespace, new_results = process_spacing(
            list(self.segments), strip_newlines
        )

        # Check for final trailing whitespace (which otherwise looks like an indent).
        if next_block and "end_of_file" in next_block.class_types and last_whitespace:
            new_results.append(
                LintResult(
                    last_whitespace,
                    [LintFix.delete(last_whitespace)],
                    description="Unnecessary trailing whitespace at end of file.",
                )
            )
            segment_buffer.remove(last_whitespace)
            last_whitespace = None

        # Is there a newline?
        # NOTE: We do this based on the segment buffer rather than self.class_types
        # because we may have just removed any present newlines in the buffer.
        if (
            any(seg.is_type("newline") for seg in segment_buffer) and not strip_newlines
        ) or (next_block and "end_of_file" in next_block.class_types):
            # Most of this section should be handled as _Indentation_.
            # BUT: There is one case we should handle here.
            # If we find that the last whitespace has a newline
            # before it, and the position markers imply there was
            # a removal between them, then remove the whitespace.
            # This ensures a consistent indent.
            if last_whitespace:
                ws_idx = self.segments.index(last_whitespace)
                if ws_idx > 0:
                    prev_seg = self.segments[ws_idx - 1]
                    if (
                        prev_seg.is_type("newline")
                        # Not just unequal. Must be actively _before_.
                        # NOTE: Based on working locations
                        and prev_seg.get_end_loc() < last_whitespace.get_start_loc()
                    ):
                        reflow_logger.debug(
                            "    Removing non-contiguous whitespace post removal."
                        )
                        segment_buffer.remove(last_whitespace)
                        # Ideally we should attach to an existing result.
                        # To do that effectively, we should look for the removed
                        # segment in the existing results.
                        temp_idx = last_whitespace.pos_marker.templated_slice.start
                        for res in existing_results:
                            if (
                                res.anchor
                                and res.anchor.pos_marker
                                and res.anchor.pos_marker.templated_slice.stop
                                == temp_idx
                            ):
                                break
                        else:  # pragma: no cover
                            raise NotImplementedError("Could not find removal result.")
                        existing_results.remove(res)
                        new_results.append(
                            LintResult(
                                res.anchor,
                                fixes=res.fixes + [LintFix("delete", last_whitespace)],
                                description=res.description,
                            )
                        )
            # Return the results.
            return existing_results + new_results, ReflowPoint(tuple(segment_buffer))

        # Otherwise this is this an inline case? (i.e. no newline)
        reflow_logger.debug(
            "    Inline case. Constraints: %s <-> %s.",
            pre_constraint,
            post_constraint,
        )

        # Do we at least have _some_ whitespace?
        if last_whitespace:
            # We do - is it the right size?
            segment_buffer, results = handle_respace__inline_with_space(
                pre_constraint,
                post_constraint,
                next_block,
                root_segment,
                segment_buffer,
                last_whitespace,
            )
            new_results.extend(results)
        else:
            # No. Should we insert some?
            # NOTE: This method operates on the existing fix buffer.
            segment_buffer, new_results, edited = handle_respace__inline_without_space(
                pre_constraint,
                post_constraint,
                prev_block,
                next_block,
                segment_buffer,
                existing_results + new_results,
                anchor_on=anchor_on,
            )
            existing_results = []
            if edited:
                reflow_logger.debug("    Modified result buffer: %s", new_results)

        # Only log if we actually made a change.
        if new_results:
            reflow_logger.debug("    New Results: %s", new_results)

        return existing_results + new_results, ReflowPoint(tuple(segment_buffer))


ReflowSequenceType = List[Union[ReflowBlock, ReflowPoint]]



"""Common classes for dialects to use."""

from typing import List, NamedTuple, Optional

from sqlfluff.core.parser import BaseSegment


class AliasInfo(NamedTuple):
    """Details about a table alias."""

    ref_str: str  # Name given to the alias
    segment: Optional[BaseSegment]  # Identifier segment containing the name
    aliased: bool
    from_expression_element: BaseSegment
    alias_expression: Optional[BaseSegment]
    object_reference: Optional[BaseSegment]


class ColumnAliasInfo(NamedTuple):
    """Details about a column alias."""

    alias_identifier_name: str
    aliased_segment: BaseSegment
    column_reference_segments: List[BaseSegment]



"""Implementation of Rule L004."""
from sqlfluff.core.parser import WhitespaceSegment
from sqlfluff.core.rules import BaseRule, LintResult, LintFix, RuleContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler
from sqlfluff.core.rules.doc_decorators import (
    document_configuration,
    document_fix_compatible,
    document_groups,
)


@document_groups
@document_fix_compatible
@document_configuration
class Rule_L004(BaseRule):
    """Incorrect indentation type.

    .. note::
       Note 1: spaces are only fixed to tabs if the number of spaces in the
       indent is an integer multiple of the ``tab_space_size`` config.

       Note 2: fixes are only applied to indents at the start of a line. Indents
       after other text on the same line are not fixed.

    **Anti-pattern**

    Using tabs instead of spaces when ``indent_unit`` config set to ``space`` (default).

    .. code-block:: sql
       :force:

        select
        ••••a,
        →   b
        from foo

    **Best practice**

    Change the line to use spaces only.

    .. code-block:: sql
       :force:

        select
        ••••a,
        ••••b
        from foo
    """

    groups = ("all", "core")
    config_keywords = ["indent_unit", "tab_space_size"]
    crawl_behaviour = SegmentSeekerCrawler({"whitespace"}, provide_raw_stack=True)

    # TODO fix indents after text:
    # https://github.com/sqlfluff/sqlfluff/pull/590#issuecomment-739484190
    def _eval(self, context: RuleContext) -> LintResult:
        """Incorrect indentation found in file."""
        # Config type hints
        self.tab_space_size: int
        self.indent_unit: str

        tab = "\t"
        space = " "
        correct_indent = self.indent
        wrong_indent = (
            tab if self.indent_unit == "space" else space * self.tab_space_size
        )
        if (
            context.segment.is_type("whitespace")
            and wrong_indent in context.segment.raw
        ):
            fixes = []
            description = "Incorrect indentation type found in file."
            edit_indent = context.segment.raw.replace(wrong_indent, correct_indent)
            pre_seg = context.raw_stack[-1] if context.raw_stack else None
            # Ensure that the number of space indents is a multiple of tab_space_size
            # before attempting to convert spaces to tabs to avoid mixed indents
            # unless we are converted tabs to spaces (indent_unit = space)
            if (
                (
                    self.indent_unit == "space"
                    or context.segment.raw.count(space) % self.tab_space_size == 0
                )
                # Only attempt a fix at the start of a newline for now
                and (pre_seg is None or pre_seg.is_type("newline"))
            ):
                fixes = [
                    LintFix.replace(
                        context.segment,
                        [
                            WhitespaceSegment(raw=edit_indent),
                        ],
                    )
                ]
            elif not (pre_seg is None or pre_seg.is_type("newline")):
                # give a helpful message if the wrong indent has been found and is not
                # at the start of a newline
                description += (
                    " The indent occurs after other text, so a manual fix is needed."
                )
            else:
                # If we get here, the indent_unit is tabs, and the number of spaces is
                # not a multiple of tab_space_size
                description += " The number of spaces is not a multiple of "
                "tab_space_size, so a manual fix is needed."
            return LintResult(
                anchor=context.segment, fixes=fixes, description=description
            )
        return LintResult()



"""Defines the plugin manager getter."""
import pluggy

from sqlfluff.core.plugin.hookspecs import PluginSpec
from sqlfluff.core.plugin import plugin_base_name, project_name


def get_plugin_manager() -> pluggy.PluginManager:
    """Initializes the PluginManager."""
    pm = pluggy.PluginManager(plugin_base_name)
    pm.add_hookspecs(PluginSpec)
    pm.load_setuptools_entrypoints(project_name)
    return pm



"""The definition of a matchable interface."""

import copy
from abc import ABC, abstractmethod
from typing import List, Optional, Tuple, TYPE_CHECKING


if TYPE_CHECKING:  # pragma: no cover
    from sqlfluff.core.parser.context import ParseContext
    from sqlfluff.core.parser.match_result import MatchResult


class Matchable(ABC):
    """A base object defining the matching interface."""

    @abstractmethod
    def is_optional(self) -> bool:
        """Return whether this element is optional."""

    @abstractmethod
    def simple(
        self, parse_context: "ParseContext", crumbs: Optional[Tuple[str, ...]] = None
    ) -> Optional[List[str]]:
        """Try to obtain a simple response from the matcher.

        NOTE: the crumbs kwarg is designed to be used by Ref to
        detect recursion.
        """

    @abstractmethod
    def match(self, segments: tuple, parse_context: "ParseContext") -> "MatchResult":
        """Match against this matcher."""

    def copy(self, **kwargs) -> "Matchable":  # pragma: no cover TODO?
        """Copy this Matchable."""
        return copy.copy(self)



"""Tests specific to the ansi dialect."""

import pytest
import logging

from sqlfluff.core import FluffConfig, Linter
from sqlfluff.core.parser import Lexer


@pytest.mark.parametrize(
    "raw,res",
    [
        # NB: The final empty string is the end of file marker.
        ("a b", ["a", " ", "b", ""]),
        ("b.c", ["b", ".", "c", ""]),
        (
            "abc \n \t def  ;blah",
            ["abc", " ", "\n", " \t ", "def", "  ", ";", "blah", ""],
        ),
    ],
)
def test__dialect__ansi__file_lex(raw, res, caplog):
    """Test we don't drop bits on simple examples."""
    config = FluffConfig(overrides=dict(dialect="ansi"))
    lexer = Lexer(config=config)
    with caplog.at_level(logging.DEBUG):
        tokens, _ = lexer.lex(raw)
    # From just the initial parse, check we're all there
    raw_list = [token.raw for token in tokens]
    assert "".join(token.raw for token in tokens) == raw
    assert raw_list == res


# Develop test to check specific elements against specific grammars.
@pytest.mark.parametrize(
    "segmentref,raw",
    [
        ("SelectKeywordSegment", "select"),
        ("NakedIdentifierSegment", "online_sales"),
        ("BareFunctionSegment", "current_timestamp"),
        ("FunctionSegment", "current_timestamp()"),
        ("NumericLiteralSegment", "1000.0"),
        ("ExpressionSegment", "online_sales / 1000.0"),
        ("IntervalExpressionSegment", "INTERVAL 1 YEAR"),
        ("ExpressionSegment", "CASE WHEN id = 1 THEN 'nothing' ELSE 'test' END"),
        # Nested Case Expressions
        # https://github.com/sqlfluff/sqlfluff/issues/172
        (
            "ExpressionSegment",
            (
                "CASE WHEN id = 1 THEN CASE WHEN true THEN 'something' "
                "ELSE 'nothing' END ELSE 'test' END"
            ),
        ),
        # Casting expressions
        # https://github.com/sqlfluff/sqlfluff/issues/161
        ("ExpressionSegment", "CAST(ROUND(online_sales / 1000.0) AS varchar)"),
        # Like expressions
        # https://github.com/sqlfluff/sqlfluff/issues/170
        ("ExpressionSegment", "name NOT LIKE '%y'"),
        # Functions with a space
        # https://github.com/sqlfluff/sqlfluff/issues/171
        ("SelectClauseElementSegment", "MIN (test.id) AS min_test_id"),
        # Interval literals
        # https://github.com/sqlfluff/sqlfluff/issues/148
        (
            "ExpressionSegment",
            "DATE_ADD(CURRENT_DATE('America/New_York'), INTERVAL 1 year)",
        ),
        # Array accessors
        ("ExpressionSegment", "my_array[1]"),
        ("ExpressionSegment", "my_array[OFFSET(1)]"),
        ("ExpressionSegment", "my_array[5:8]"),
        ("ExpressionSegment", "4 + my_array[OFFSET(1)]"),
        ("ExpressionSegment", "bits[OFFSET(0)] + 7"),
        (
            "SelectClauseElementSegment",
            (
                "(count_18_24 * bits[OFFSET(0)])"
                " / audience_size AS relative_abundance"
            ),
        ),
        ("ExpressionSegment", "count_18_24 * bits[OFFSET(0)] + count_25_34"),
        (
            "SelectClauseElementSegment",
            (
                "(count_18_24 * bits[OFFSET(0)] + count_25_34)"
                " / audience_size AS relative_abundance"
            ),
        ),
        # Dense math expressions
        # https://github.com/sqlfluff/sqlfluff/issues/178
        # https://github.com/sqlfluff/sqlfluff/issues/179
        ("SelectStatementSegment", "SELECT t.val/t.id FROM test WHERE id*1.0/id > 0.8"),
        ("SelectClauseElementSegment", "t.val/t.id"),
        # Issue with casting raise as part of PR #177
        ("SelectClauseElementSegment", "CAST(num AS INT64)"),
        # Casting as datatype with arguments
        ("SelectClauseElementSegment", "CAST(num AS numeric(8,4))"),
        # Wildcard field selection
        ("SelectClauseElementSegment", "a.*"),
        ("SelectClauseElementSegment", "a.b.*"),
        ("SelectClauseElementSegment", "a.b.c.*"),
        # Default Element Syntax
        ("ObjectReferenceSegment", "a..c.*"),
        # Negative Elements
        ("SelectClauseElementSegment", "-some_variable"),
        ("SelectClauseElementSegment", "- some_variable"),
        # Complex Functions
        (
            "ExpressionSegment",
            "concat(left(uaid, 2), '|', right(concat('0000000', "
            "SPLIT_PART(uaid, '|', 4)), 10), '|', '00000000')",
        ),
        # Notnull and Isnull
        ("ExpressionSegment", "c is null"),
        ("ExpressionSegment", "c is not null"),
        ("SelectClauseElementSegment", "c is null as c_isnull"),
        ("SelectClauseElementSegment", "c is not null as c_notnull"),
        # Shorthand casting
        ("ExpressionSegment", "NULL::INT"),
        ("SelectClauseElementSegment", "NULL::INT AS user_id"),
        ("TruncateStatementSegment", "TRUNCATE TABLE test"),
        ("TruncateStatementSegment", "TRUNCATE test"),
    ],
)
def test__dialect__ansi_specific_segment_parses(
    segmentref, raw, caplog, dialect_specific_segment_parses
):
    """Test that specific segments parse as expected.

    NB: We're testing the PARSE function not the MATCH function
    although this will be a recursive parse and so the match
    function of SUBSECTIONS will be tested if present. The match
    function of the parent will not be tested.
    """
    dialect_specific_segment_parses("ansi", segmentref, raw, caplog)


@pytest.mark.parametrize(
    "segmentref,raw",
    [
        # Check we don't match empty whitespace as a reference
        ("ObjectReferenceSegment", "\n     ")
    ],
)
def test__dialect__ansi_specific_segment_not_match(
    segmentref, raw, caplog, dialect_specific_segment_not_match
):
    """Test that specific segments do not match.

    NB: We're testing the MATCH function not the PARSE function.
    This is the opposite to the above.
    """
    dialect_specific_segment_not_match("ansi", segmentref, raw, caplog)


@pytest.mark.parametrize(
    "raw,err_locations",
    [
        # Missing Closing bracket. Error should be raised
        # on the starting bracket.
        ("SELECT 1 + (2 ", [(1, 12)]),
        # Set expression with inappropriate ORDER BY or LIMIT. Error
        # raised on the UNION.
        ("SELECT * FROM a ORDER BY 1 UNION SELECT * FROM b", [(1, 27)]),
        ("SELECT * FROM a LIMIT 1 UNION SELECT * FROM b", [(1, 24)]),
        ("SELECT * FROM a ORDER BY 1 LIMIT 1 UNION SELECT * FROM b", [(1, 35)]),
    ],
)
def test__dialect__ansi_specific_segment_not_parse(raw, err_locations, caplog):
    """Test queries do not parse, with parsing errors raised properly."""
    lnt = Linter(dialect="ansi")
    parsed = lnt.parse_string(raw)
    assert len(parsed.violations) > 0
    print(parsed.violations)
    locs = [(v.line_no, v.line_pos) for v in parsed.violations]
    assert locs == err_locations


def test__dialect__ansi_is_whitespace():
    """Test proper tagging with is_whitespace."""
    lnt = Linter(dialect="ansi")
    with open("test/fixtures/dialects/ansi/select_in_multiline_comment.sql") as f:
        parsed = lnt.parse_string(f.read())
    # Check all the segments that *should* be whitespace, ARE
    for raw_seg in parsed.tree.get_raw_segments():
        if raw_seg.is_type("whitespace", "newline"):
            assert raw_seg.is_whitespace


@pytest.mark.parametrize(
    "sql_string, indented_joins,meta_loc",
    [
        ("select field_1 from my_table as alias_1", True, (1, 5, 8, 14, 15)),
        ("select field_1 from my_table as alias_1", False, (1, 5, 8, 14, 15)),
        (
            "select field_1 from my_table as alias_1 join foo using (field_1)",
            True,
            (1, 5, 8, 16, 21, 24, 26, 28, 29, 30, 31),
        ),
        (
            "select field_1 from my_table as alias_1 join foo using (field_1)",
            False,
            (1, 5, 8, 15, 17, 22, 25, 27, 29, 30, 31),
        ),
    ],
)
def test__dialect__ansi_parse_indented_joins(sql_string, indented_joins, meta_loc):
    """Test parsing of meta segments using Conditional works with indented_joins."""
    lnt = Linter(
        config=FluffConfig(
            configs={"indentation": {"indented_joins": indented_joins}},
            overrides={"dialect": "ansi"},
        )
    )
    parsed = lnt.parse_string(sql_string)
    # Check that there's nothing unparsable
    assert "unparsable" not in parsed.tree.type_set()
    # Check all the segments that *should* be metas, ARE.
    # NOTE: This includes the end of file marker.
    res_meta_locs = tuple(
        idx
        for idx, raw_seg in enumerate(parsed.tree.get_raw_segments())
        if raw_seg.is_meta
    )
    assert res_meta_locs == meta_loc



"""A list of all SQL key words."""

redshift_reserved_keywords = """AES128
AES256
ALL
ALLOWOVERWRITE
ANALYSE
ANALYZE
AND
ANY
ARRAY
AS
ASC
AUTHORIZATION
AZ64
BETWEEN
BINARY
BLANKSASNULL
BOTH
BYTEDICT
CASE
CAST
CHECK
COLLATE
COLUMN
COMPROWS
COMPUPDATE
CONSTRAINT
CREATE
CREDENTIALS
CROSS
CURRENT_DATE
CURRENT_TIME
CURRENT_TIMESTAMP
CURRENT_USER
CURRENT_USER_ID
DATETIME
DEFAULT
DEFERRABLE
DEFRAG
DELIMITERS
DELTA
DELTA32K
DESC
DISABLE
DISTINCT
DO
ELSE
EMPTYASNULL
ENABLE
ENCRYPT
ENCRYPTION
END
EXCEPT
EXPLICIT_IDS
FALSE
FILLRECORD
FOR
FOREIGN
FREEZE
FROM
FULL
GLOBALDICT256
GLOBALDICT64K
GRANT
GROUP
HAVING
IDENTITY
IGNORE
IGNOREBLANKLINES
IGNOREHEADER
ILIKE
IN
INITIALLY
INNER
INTERSECT
INTO
IS
ISNULL
JOIN
LEADING
LEFT
LIKE
LIMIT
LOCALTIME
LOCALTIMESTAMP
LUN
LUNS
LZO
MINUS
MOSTLY16
MOSTLY32
MOSTLY8
NATURAL
NEW
NOT
NOTNULL
NULL
NULLS
OFF
OFFSET
OID
OLD
ON
ONLY
OPEN
OR
ORDER
OUTER
OVERLAPS
PARALLEL
PARTITION
PERCENT
PERMISSIONS
PIVOT
PLACING
PRIMARY
RAW
READRATIO
RECOVER
REFERENCES
RESPECT
REJECTLOG
RESORT
RESTORE
RIGHT
RUNLENGTH
SELECT
SESSION_USER
SIMILAR
SNAPSHOT
SOME
SYSDATE
SYSTEM
TABLE
TAG
TDES
TEXT255
TEXT32K
THEN
TIMESTAMP
TO
TOP
TRAILING
TRUE
TRUNCATECOLUMNS
UNION
UNIQUE
UNNEST
UNPIVOT
USER
USING
VERBOSE
WHEN
WHERE
WITH
WITHIN
WITHOUT"""

redshift_unreserved_keywords = """A
ABORT
ABS
ABSENT
ABSOLUTE
ACCEPTANYDATE
ACCEPTINVCHARS
ACCESS
ACCESS_KEY_ID
ACCORDING
ACCOUNT
ACOS
ACTION
ADA
ADD
ADDQUOTES
ADMIN
AFTER
AGGREGATE
ALLOCATE
ALSO
ALTER
ALWAYS
ANYELEMENT
APPLY
APPROXIMATE
ARE
ARRAY_AGG
ARRAY_MAX_CARDINALITY
ASENSITIVE
ASIN
ASSERTION
ASSIGNMENT
ASYMMETRIC
AT
ATAN
ATOMIC
ATTACH
ATTRIBUTE
ATTRIBUTES
AUTO
AUTO_INCREMENT
AVG
AVRO
BACKUP
BACKWARD
BASE64
BEFORE
BEGIN
BEGIN_FRAME
BEGIN_PARTITION
BERNOULLI
BIGINT
BINARY_CLASSIFICATION
BINDING
BIT
BIT_LENGTH
BLANKSASNULL
BLOB
BLOCKED
BOM
BOOL
BOOLEAN
BOOST
BPCHAR
BREADTH
BUFFERS
BY
BYPASSRLS
BZIP2
C
CACHE
CALL
CALLED
CARDINALITY
CASCADE
CASCADED
CASE_INSENSITIVE
CASE_SENSITIVE
CATALOG
CATALOG_NAME
CATALOG_ROLE
CEIL
CEILING
CHAIN
CHAINING
CHAR
CHARACTER
CHARACTERISTICS
CHARACTERS
CHARACTER_LENGTH
CHARACTER_SET_CATALOG
CHARACTER_SET_NAME
CHARACTER_SET_SCHEMA
CHAR_LENGTH
CHECKPOINT
CLASS
CLASSIFIER
CLASS_ORIGIN
CLEANPATH
CLOB
CLOSE
CLUSTER
COALESCE
COBOL
COLLATION
COLLATION_CATALOG
COLLATION_NAME
COLLATION_SCHEMA
COLLECT
COLUMNS
COLUMN_NAME
COMMAND_FUNCTION
COMMAND_FUNCTION_CODE
COMMENT
COMMENTS
COMMIT
COMMITTED
COMPOUND
COMPRESSION
CONCURRENTLY
CONDITION
CONDITIONAL
CONDITION_NUMBER
CONFIGURATION
CONFLICT
CONNECT
CONNECTION
CONNECTION_NAME
CONSTRAINTS
CONSTRAINT_CATALOG
CONSTRAINT_NAME
CONSTRAINT_SCHEMA
CONSTRUCTOR
CONTAINS
CONTENT
CONTINUE
CONTROL
CONVERSION
CONVERT
COPY
CORR
CORRESPONDING
COS
COSH
COST
COSTS
COUNT
COVAR_POP
COVAR_SAMP
CREATEDB
CREATEUSER
CREATEROLE
CSV
CUBE
CUME_DIST
CURRENT
CURRENT_CATALOG
CURRENT_DEFAULT_TRANSFORM_GROUP
CURRENT_PATH
CURRENT_ROLE
CURRENT_ROW
CURRENT_SCHEMA
CURRENT_TRANSFORM_GROUP_FOR_TYPE
CURSOR
CURSOR_NAME
CYCLE
DATA
DATABASE
DATALINK
DATASHARE
DATASHARES
DATE
DATEFORMAT
DATETIME_INTERVAL_CODE
DATETIME_INTERVAL_PRECISION
DAY
DAYOFYEAR
DB
DEALLOCATE
DEC
DECFLOAT
DECIMAL
DECLARE
DEFAULTS
DEFERRED
DEFINE
DEFINED
DEFINER
DEFLATE
DEGREE
DELETE
DELIMITED
DELIMITER
DENSE_RANK
DEPENDS
DEPTH
DEREF
DERIVED
DESCRIBE
DESCRIPTOR
DETACH
DETERMINISTIC
DIAGNOSTICS
DICTIONARY
DISCARD
DISCONNECT
DISPATCH
DISTKEY
DISTSTYLE
DLNEWCOPY
DLPREVIOUSCOPY
DLURLCOMPLETE
DLURLCOMPLETEONLY
DLURLCOMPLETEWRITE
DLURLPATH
DLURLPATHONLY
DLURLPATHWRITE
DLURLSCHEME
DLURLSERVER
DLVALUE
DOCUMENT
DOMAIN
DOUBLE
DROP
DYNAMIC
DYNAMIC_FUNCTION
DYNAMIC_FUNCTION_CODE
EACH
ELEMENT
EMPTY
ENCODE
ENCODING
ENCRYPTED
END-EXEC
END_FRAME
END_PARTITION
ENFORCED
ENUM
EPOCH
EPOCHSECS
EPOCHMILLISECS
EQUALS
ERROR
ESCAPE
EVEN
EVENT
EVERY
EXCEPTION
EXCLUDE
EXCLUDING
EXCLUSIVE
EXEC
EXECUTE
EXECUTION
EXISTS
EXP
EXPLAIN
EXPLICIT
EXPRESSION
EXTENDED
EXTENSION
EXTERNAL
EXTRACT
FAMILY
FETCH
FIELDS
FILE
FILTER
FINAL
FINALIZE
FINISH
FIRST
FIRST_VALUE
FIXEDWIDTH
FLAG
FLOAT
FLOAT4
FLOAT8
FLOOR
FOLLOWING
FORCE
FORMAT
FORTRAN
FORWARD
FOUND
FRAME_ROW
FREE
FS
FULFILL
FUNCTION
FUNCTIONS
FUSION
FUTURE
G
GB
GENERAL
GENERATED
GEOGRAPHY
GEOMETRY
GET
GLOBAL
GO
GOTO
GRANTED
GRANTS
GREATEST
GROUPING
GROUPS
GZIP
HANDLER
HASH
HEADER
HEX
HIERARCHY
HIVE
HLLSKETCH
HOLD
HOUR
HYPERPARAMETERS
IAM_ROLE
ID
IF
IMMEDIATE
IMMEDIATELY
IMMUTABLE
IMPLEMENTATION
IMPLICIT
IMPORT
IMPORTED
INCLUDE
INCLUDENEW
INCLUDING
INCREMENT
INDENT
INDEX
INDEXES
INDICATOR
INHERIT
INHERITS
INITIAL
INLINE
INOUT
INPUT
INPUTFORMAT
INSENSITIVE
INSERT
INSTANCE
INSTANTIABLE
INSTEAD
INT
INT2
INT4
INT8
INTEGER
INTEGRATION
INTEGRITY
INTERLEAVED
INTERSECTION
INTERVAL
INVOKER
ISOLATION
JSON
JSON_ARRAY
JSON_ARRAYAGG
JSON_EXISTS
JSON_OBJECT
JSON_OBJECTAGG
JSON_QUERY
JSON_TABLE
JSON_TABLE_PRIMITIVE
JSON_VALUE
K
KEEP
KEY
KEYS
KEY_MEMBER
KEY_TYPE
KINESIS
KMEANS
KMS_KEY_ID
LABEL
LAG
LAMBDA
LANGUAGE
LARGE
LAST
LAST_VALUE
LATERAL
LEAD
LEAKPROOF
LEAST
LENGTH
LEVEL
LIBRARY
LIKE_REGEX
LINES
LINK
LIST
LISTAGG
LISTEN
LN
LOAD
LOCAL
LOCATION
LOCATOR
LOCK
LOCKED
LOG
LOG10
LOGGED
LOGIN
LOWER
LZOP
M
MAIN
MANAGE
MANIFEST
MAP
MAPPING
MASKING
MASTER_SYMMETRIC_KEY
MATCH
MATCHED
MATCHES
MATCH_NUMBER
MATCH_RECOGNIZE
MATERIALIZED
MAX
MAXERROR
MAXFILESIZE
MAXVALUE
MAX_CELLS
MAX_RUNTIME
MB
MEASURES
MEMBER
MERGE
MESSAGE_LENGTH
MESSAGE_OCTET_LENGTH
MESSAGE_TEXT
METASTORE
METHOD
MILLISECOND
MIN
MINUTE
MINVALUE
ML
MLP
MOD
MODE
MODEL
MODEL_TYPE
MODIFIES
MODIFY
MODULE
MODULUS
MONITOR
MONTH
MORE
MOVE
MULTICLASS_CLASSIFICATION
MULTISET
MYSQL
MUMPS
NAME
NAMES
NAMESPACE
NAN
NATIONAL
NCHAR
NCLOB
NESTED
NESTING
NEXT
NFC
NFD
NFKC
NFKD
NIL
NO
NOBYPASSRLS
NOCACHE
NOCREATEDB
NOCREATEROLE
NOCREATEUSER
NOCYCLE
NOINHERIT
NOLOAD
NOLOGIN
NOREPLICATION
NOSUPERUSER
NONE
NOORDER
NORMALIZE
OUTPUTFORMAT
NORMALIZED
NOTHING
NOTIFY
NOWAIT
NTH_VALUE
NTILE
NULLABLE
NULLIF
NUMBER
NUMERIC
NVARCHAR
OBJECT
OBJECTIVE
OCCURRENCES_REGEX
OCTET_LENGTH
OCTETS
OF
OFFLINE
OIDS
OMIT
ONE
OPERATE
OPERATOR
OPTION
OPTIONS
ORC
ORDERING
ORDINALITY
OTHERS
OUT
OUTPUT
OVER
OVERFLOW
OVERLAY
OVERRIDING
OVERWRITE
OWNED
OWNER
OWNERSHIP
P
PAD
PARAMETER
PARAMETER_MODE
PARAMETER_NAME
PARAMETER_ORDINAL_POSITION
PARAMETER_SPECIFIC_CATALOG
PARAMETER_SPECIFIC_NAME
PARAMETER_SPECIFIC_SCHEMA
PARQUET
PARSER
PARTIAL
PARTITIONED
PASCAL
PASS
PASSING
PASSTHROUGH
PASSWORD
PAST
PATH
PATTERN
PER
PERCENT_RANK
PERCENTILE_CONT
PERCENTILE_DISC
PERIOD
PERMISSION
PERMUTE
PIPE
PLAIN
PLAN
PLANS
PLI
POLICY
PORT
PORTION
POSITION
POSITION_REGEX
POSTGRES
POWER
PRECEDES
PRECEDING
PRECISION
PREPARE
PREPARED
PREPROCESSORS
PRESERVE
PRESET
PRIOR
PRIVATE
PRIVILEGES
PROBLEM_TYPE
PROCEDURAL
PROCEDURE
PROCEDURES
PROGRAM
PROPERTIES
PRUNE
PTF
PUBLIC
PUBLICACCESSIBLE
PUBLICATION
PLPYTHONU
QUALIFY
QUARTER
QUOTA
QUOTE
QUOTES
RANGE
RANK
RCFILE
READ
READRATIO
READS
REAL
REASSIGN
RECHECK
RECLUSTER
RECOVERY
RECURSIVE
REDSHIFT
REF
REFCURSOR
REFERENCE_USAGE
REFERENCING
REFRESH
REGION
REGR_AVGX
REGR_AVGY
REGR_COUNT
REGR_INTERCEPT
REGR_R2
REGR_SLOPE
REGR_SXX
REGR_SXY
REGR_SYY
REGRESSION
REINDEX
RELATIVE
RELEASE
REMAINDER
REMOVE
REMOVEQUOTES
RENAME
REPEATABLE
REPLACE
REPLICA
REPLICATION
REQUIRING
RESET
RESOURCE
RESTART
RESTRICT
RESTRICTED
RESULT
RETRY_TIMEOUT
RETURN
RETURNED_CARDINALITY
RETURNED_LENGTH
RETURNED_OCTET_LENGTH
RETURNED_SQLSTATE
RETURNING
RETURNS
REVOKE
RLIKE
RLS
ROLE
ROLLBACK
ROLLUP
ROUNDEC
ROUTINE
ROUTINE_CATALOG
ROUTINE_NAME
ROUTINE_SCHEMA
ROUTINES
ROW
ROW_COUNT
ROW_NUMBER
ROWGROUPSIZE
ROWS
RULE
RUNNING
S3_BUCKET
S3_GARBAGE_COLLECT
SAFE
SAGEMAKER
SAVEPOINT
SCALAR
SCALE
SCHEMA
SCHEMA_NAME
SCHEMAS
SCOPE
SCOPE_CATALOG
SCOPE_NAME
SCOPE_SCHEMA
SCROLL
SEARCH
SECOND
SECRET_ACCESS_KEY
SECRET_ARN
SECTION
SECURITY
SEEK
SELECTIVE
SELF
SENSITIVE
SEPARATOR
SEQUENCE
SEQUENCEFILE
SEQUENCES
SERDE
SERDEPROPERTIES
SERIALIZABLE
SERVER
SERVER_NAME
SESSION
SESSION_TOKEN
SET
SETTINGS
SETOF
SETS
SHAPEFILE
SHARE
SHOW
SIMPLE
SIMPLIFY
SIN
SINH
SIZE
SKIP
SMALLINT
SORT
SORTKEY
SOURCE
SPACE
SPECIFIC
SPECIFIC_NAME
SPECIFICTYPE
SQL
SQLCODE
SQLERROR
SQLEXCEPTION
SQLSTATE
SQLWARNING
SQRT
STABLE
STAGE
STAGES
STANDALONE
START
STATE
STATEMENT
STATIC
STATISTICS
STATUPDATE
STDDEV_POP
STDDEV_SAMP
STDIN
STDOUT
STORAGE
STORED
STREAM
STREAMS
STRICT
STRING
STRIP
STRUCTURE
STYLE
SUBCLASS_ORIGIN
SUBMULTISET
SUBSCRIPTION
SUBSET
SUBSTRING
SUBSTRING_REGEX
SUCCEEDS
SUM
SUPER
SUPERUSER
SUPPORT
SYMMETRIC
SYSID
SYSLOG
SYSTEM_TIME
SYSTEM_USER
T
TABLE_NAME
TABLES
TABLESAMPLE
TABLESPACE
TAN
TANH
TARGET
TASK
TASKS
TB
TEMP
TEMPLATE
TEMPORARY
TERMINATED
TEXT
TEXTFILE
THROUGH
TIES
TIME
TIMEFORMAT
TIMEOUT
TIMETZ
TIMESTAMPTZ
TIMEZONE_HOUR
TIMEZONE_MINUTE
TOKEN
TOP_LEVEL_COUNT
TRANSACTION
TRANSACTION_ACTIVE
TRANSACTIONS_COMMITTED
TRANSACTIONS_ROLLED_BACK
TRANSFORM
TRANSFORMS
TRANSIENT
TRANSLATE
TRANSLATE_REGEX
TRANSLATION
TREAT
TRIGGER
TRIGGER_CATALOG
TRIGGER_NAME
TRIGGER_SCHEMA
TRIM
TRIMBLANKS
TRIM_ARRAY
TRUNCATE
TRUNCATECOLUMNS
TRUSTED
TYPE
TYPES
UESCAPE
UNBOUNDED
UNCOMMITTED
UNCONDITIONAL
UNDER
UNENCRYPTED
UNKNOWN
UNLIMITED
UNLINK
UNLISTEN
UNLOAD
UNLOGGED
UNMATCHED
UNNAMED
UNRESTRICTED
UNSAFE
UNSIGNED
UNTIL
UNTYPED
UPDATE
UPPER
URI
USE_ANY_ROLE
USAGE
USE
USER_DEFINED_TYPE_CATALOG
USER_DEFINED_TYPE_CODE
USER_DEFINED_TYPE_NAME
USER_DEFINED_TYPE_SCHEMA
UTF16
UTF16BE
UTF16LE
UTF32
UTF8
VACUUM
VALID
VALIDATE
VALIDATOR
VALUE
VALUE_OF
VALUES
VAR_POP
VAR_SAMP
VARBINARY
VARBYTE
VARCHAR
VARIADIC
VARYING
VERSION
VERSIONING
VIEW
VIEWS
VOLATILE
WALLET
WAREHOUSE
WEEK
WEEKDAY
WHENEVER
WHITESPACE
WIDTH_BUCKET
WINDOW
WORK
WRAPPER
WRITE
XGBOOST
XML
XMLAGG
XMLATTRIBUTES
XMLBINARY
XMLCAST
XMLCOMMENT
XMLCONCAT
XMLDECLARATION
XMLDOCUMENT
XMLELEMENT
XMLEXISTS
XMLFOREST
XMLITERATE
XMLNAMESPACES
XMLPARSE
XMLPI
XMLQUERY
XMLROOT
XMLSCHEMA
XMLSERIALIZE
XMLTABLE
XMLTEXT
XMLVALIDATE
YAML
YEAR
YES
ZONE
ZSTD"""



"""Implementation of Rule L002."""
from typing import Optional

from sqlfluff.core.rules import BaseRule, LintResult, LintFix, RuleContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler
from sqlfluff.core.rules.doc_decorators import (
    document_configuration,
    document_fix_compatible,
    document_groups,
)


@document_groups
@document_fix_compatible
@document_configuration
class Rule_L002(BaseRule):
    """Mixed Tabs and Spaces in single whitespace.

    This rule will fail if a single section of whitespace
    contains both tabs and spaces.

    **Anti-pattern**

    The ``•`` character represents a space and the ``→`` character represents a tab.
    In this example, the second line contains two spaces and one tab.

    .. code-block:: sql
       :force:

        SELECT
        ••→a
        FROM foo

    **Best practice**

    Change the line to use spaces only.

    .. code-block:: sql
       :force:

        SELECT
        ••••a
        FROM foo

    """

    groups = ("all", "core")
    config_keywords = ["tab_space_size"]
    crawl_behaviour = SegmentSeekerCrawler({"whitespace"}, provide_raw_stack=True)

    def _eval(self, context: RuleContext) -> Optional[LintResult]:
        """Mixed Tabs and Spaces in single whitespace.

        Only trigger from whitespace segments if they contain
        multiple kinds of whitespace.
        """
        # Config type hints
        self.tab_space_size: int

        if context.segment.is_type("whitespace"):
            if " " in context.segment.raw and "\t" in context.segment.raw:
                if not context.raw_stack or context.raw_stack[-1].is_type("newline"):
                    # We've got a single whitespace at the beginning of a line.
                    # It's got a mix of spaces and tabs. Replace each tab with
                    # a multiple of spaces
                    return LintResult(
                        anchor=context.segment,
                        fixes=[
                            LintFix.replace(
                                context.segment,
                                [
                                    context.segment.edit(
                                        context.segment.raw.replace(
                                            "\t", " " * self.tab_space_size
                                        )
                                    ),
                                ],
                            ),
                        ],
                    )
        return None



"""Defines commonly used segment predicates for rule writers.

For consistency, all the predicates in this module are implemented as functions
returning functions. This avoids rule writers having to remember the
distinction between normal functions and functions returning functions.

This is not necessarily a complete set of predicates covering all possible
requirements. Rule authors can define their own predicates as needed, either
as regular functions, `lambda`, etc.
"""
from typing import Callable, Optional

from sqlfluff.core.parser import BaseSegment
from sqlfluff.utils.functional.raw_file_slices import RawFileSlices
from sqlfluff.utils.functional.templated_file_slices import TemplatedFileSlices
from sqlfluff.core.templaters.base import (
    RawFileSlice,
    TemplatedFile,
)


def raw_is(*raws: str) -> Callable[[BaseSegment], bool]:  # pragma: no cover
    """Returns a function that determines if segment matches one of the raw inputs."""

    def _(segment: BaseSegment):
        return segment.raw in raws

    return _


def raw_upper_is(*raws: str) -> Callable[[BaseSegment], bool]:
    """Returns a function that determines if segment matches one of the raw inputs."""

    def _(segment: BaseSegment):
        return segment.raw_upper in raws

    return _


def is_type(*seg_type: str) -> Callable[[BaseSegment], bool]:
    """Returns a function that determines if segment is one of the types."""

    def _(segment: BaseSegment):
        return segment.is_type(*seg_type)

    return _


def is_keyword(*keyword_name) -> Callable[[BaseSegment], bool]:
    """Returns a function that determines if it's a matching keyword."""
    return and_(
        is_type("keyword"), raw_upper_is(*[raw.upper() for raw in keyword_name])
    )


def is_code() -> Callable[[BaseSegment], bool]:
    """Returns a function that checks if segment is code."""

    def _(segment: BaseSegment) -> bool:
        return segment.is_code

    return _


def is_comment() -> Callable[[BaseSegment], bool]:
    """Returns a function that checks if segment is comment."""

    def _(segment: BaseSegment) -> bool:
        return segment.is_comment

    return _


def is_expandable() -> Callable[[BaseSegment], bool]:
    """Returns a function that checks if segment is expandable."""

    def _(segment: BaseSegment) -> bool:
        return segment.is_expandable

    return _


def is_meta() -> Callable[[BaseSegment], bool]:
    """Returns a function that checks if segment is meta."""

    def _(segment: BaseSegment) -> bool:
        return segment.is_meta

    return _


def is_raw() -> Callable[[BaseSegment], bool]:
    """Returns a function that checks if segment is raw."""

    def _(segment: BaseSegment) -> bool:
        return segment.is_raw()

    return _


def is_whitespace() -> Callable[[BaseSegment], bool]:
    """Returns a function that checks if segment is whitespace."""

    def _(segment: BaseSegment) -> bool:
        return segment.is_whitespace

    return _


def is_templated() -> Callable[[BaseSegment], bool]:  # pragma: no cover
    """Returns a function that checks if segment is templated."""

    def _(segment: BaseSegment) -> bool:
        return segment.is_templated

    return _


def get_type() -> Callable[[BaseSegment], str]:
    """Returns a function that gets segment type."""

    def _(segment: BaseSegment) -> str:
        return segment.get_type()

    return _


def and_(*functions: Callable[[BaseSegment], bool]) -> Callable[[BaseSegment], bool]:
    """Returns a function that computes the functions and-ed together."""

    def _(segment: BaseSegment):
        return all(function(segment) for function in functions)

    return _


def or_(*functions: Callable[[BaseSegment], bool]) -> Callable[[BaseSegment], bool]:
    """Returns a function that computes the functions or-ed together."""

    def _(segment: BaseSegment):
        return any(function(segment) for function in functions)

    return _


def not_(fn: Callable[[BaseSegment], bool]) -> Callable[[BaseSegment], bool]:
    """Returns a function that computes: not fn()."""

    def _(segment: BaseSegment):
        return not fn(segment)

    return _


def raw_slices(
    segment: BaseSegment,
    templated_file: Optional[TemplatedFile],
) -> RawFileSlices:  # pragma: no cover
    """Returns raw slices for a segment."""
    if not templated_file:
        raise ValueError(
            'raw_slices: "templated_file" parameter is required.'
        )  # pragma: no cover
    if not segment.pos_marker:
        raise ValueError(
            'raw_slices: "segment" parameter must have pos_marker set.'
        )  # pragma: no cover
    return RawFileSlices(
        *templated_file.raw_slices_spanning_source_slice(
            segment.pos_marker.source_slice
        ),
        templated_file=templated_file
    )


def templated_slices(
    segment: BaseSegment,
    templated_file: Optional[TemplatedFile],
) -> TemplatedFileSlices:
    """Returns raw slices for a segment."""
    if not templated_file:
        raise ValueError(
            'templated_slices: "templated_file" parameter is required.'
        )  # pragma: no cover
    if not segment.pos_marker:
        raise ValueError(
            'templated_slices: "segment" parameter must have pos_marker set.'
        )  # pragma: no cover
    # :TRICKY: We don't use _find_slice_indices_of_templated_pos() here because
    # it treats TemplatedFileSlice.templated_slice.stop as inclusive, not
    # exclusive. Other parts of SQLFluff rely on this behavior, but we don't
    # want it. It's easy enough to do this ourselves.
    start = segment.pos_marker.templated_slice.start
    stop = segment.pos_marker.templated_slice.stop
    templated_slices = [
        slice_
        for slice_ in templated_file.sliced_file
        if (stop > slice_.templated_slice.start and start < slice_.templated_slice.stop)
    ]
    return TemplatedFileSlices(*templated_slices, templated_file=templated_file)


def raw_slice(segment: BaseSegment, raw_slice_: RawFileSlice) -> str:
    """Return the portion of a segment's source provided by raw_slice."""
    result = ""
    if not segment.pos_marker:
        raise ValueError(
            'raw_slice: "segment" parameter must have pos_marker set.'
        )  # pragma: no cover
    seg_start = segment.pos_marker.source_slice.start
    seg_stop = segment.pos_marker.source_slice.stop
    if seg_start != seg_stop:
        start = max(seg_start, raw_slice_.source_idx)
        stop = min(
            seg_stop,
            raw_slice_.source_idx + len(raw_slice_.raw),
        )
        result = segment.pos_marker.templated_file.source_str[slice(start, stop)]
    return result



"""Sequence and Bracketed Grammars."""

from typing import Optional, List, Tuple, cast

from sqlfluff.core.errors import SQLParseError

from sqlfluff.core.parser.segments import (
    BaseSegment,
    Indent,
    Dedent,
    allow_ephemeral,
    BracketedSegment,
    MetaSegment,
)
from sqlfluff.core.parser.helpers import trim_non_code_segments, check_still_complete
from sqlfluff.core.parser.match_result import MatchResult
from sqlfluff.core.parser.match_wrapper import match_wrapper
from sqlfluff.core.parser.context import ParseContext
from sqlfluff.core.parser.grammar.base import (
    BaseGrammar,
    cached_method_for_parse_context,
)
from sqlfluff.core.parser.grammar.conditional import Conditional
from os import getenv


class Sequence(BaseGrammar):
    """Match a specific sequence of elements."""

    test_env = getenv("SQLFLUFF_TESTENV", "")

    @cached_method_for_parse_context
    def simple(self, parse_context: ParseContext, crumbs=None) -> Optional[List[str]]:
        """Does this matcher support a uppercase hash matching route?

        Sequence does provide this, as long as the *first* non-optional
        element does, *AND* and optional elements which preceded it also do.
        """
        simple_buff = []
        for opt in self._elements:
            simple = opt.simple(parse_context=parse_context, crumbs=crumbs)
            if not simple:
                return None
            simple_buff += simple

            if not opt.is_optional():
                # We found our first non-optional element!
                return simple_buff
        # If *all* elements are optional AND simple, I guess it's also simple.
        return simple_buff

    @match_wrapper()
    @allow_ephemeral
    def match(self, segments, parse_context):
        """Match a specific sequence of elements."""
        if isinstance(segments, BaseSegment):
            segments = tuple(segments)  # pragma: no cover TODO?

        matched_segments = MatchResult.from_empty()
        unmatched_segments = segments

        # Buffers of uninstantiated meta segments.
        meta_pre_nc = ()
        meta_post_nc = ()
        early_break = False

        for idx, elem in enumerate(self._elements):
            # Check for an early break.
            if early_break:
                break

            while True:
                # Consume non-code if appropriate
                if self.allow_gaps:
                    pre_nc, mid_seg, post_nc = trim_non_code_segments(
                        unmatched_segments
                    )
                else:
                    pre_nc = ()
                    mid_seg = unmatched_segments
                    post_nc = ()

                # Is it an indent or dedent?
                if elem.is_meta:
                    # Elements with a negative indent value come AFTER
                    # the whitespace. Positive or neutral come BEFORE.
                    if elem.indent_val < 0:
                        meta_post_nc += (elem(),)
                    else:
                        meta_pre_nc += (elem(),)
                    break

                # Is it a conditional? If so is it active
                if isinstance(elem, Conditional) and not elem.is_enabled(parse_context):
                    # If it's not active, skip it.
                    break

                if len(pre_nc + mid_seg + post_nc) == 0:
                    # We've run our of sequence without matching everything.
                    # Do only optional or meta elements remain?
                    if all(
                        e.is_optional() or e.is_meta or isinstance(e, Conditional)
                        for e in self._elements[idx:]
                    ):
                        # then it's ok, and we can return what we've got so far.
                        # No need to deal with anything left over because we're at the
                        # end, unless it's a meta segment.

                        # We'll add those meta segments after any existing ones. So
                        # the go on the meta_post_nc stack.
                        for e in self._elements[idx:]:
                            # If it's meta, instantiate it.
                            if e.is_meta:
                                meta_post_nc += (e(),)  # pragma: no cover TODO?
                            # If it's conditional and it's enabled, match it.
                            if isinstance(e, Conditional) and e.is_enabled(
                                parse_context
                            ):
                                meta_match = e.match(tuple(), parse_context)
                                if meta_match:
                                    meta_post_nc += meta_match.matched_segments

                        # Early break to exit via the happy match path.
                        early_break = True
                        break
                    else:
                        # we've got to the end of the sequence without matching all
                        # required elements.
                        return MatchResult.from_unmatched(segments)
                else:
                    # We've already dealt with potential whitespace above, so carry on
                    # to matching
                    with parse_context.deeper_match() as ctx:
                        elem_match = elem.match(mid_seg, parse_context=ctx)

                    if elem_match.has_match():
                        # We're expecting mostly partial matches here, but complete
                        # matches are possible. Don't be greedy with whitespace!
                        matched_segments += (
                            meta_pre_nc
                            + pre_nc
                            + meta_post_nc
                            + elem_match.matched_segments
                        )
                        meta_pre_nc = ()
                        meta_post_nc = ()
                        unmatched_segments = elem_match.unmatched_segments + post_nc
                        # Each time we do this, we do a sense check to make sure we
                        # haven't dropped anything. (Because it's happened before!).
                        if self.test_env:
                            check_still_complete(
                                segments,
                                matched_segments.matched_segments,
                                unmatched_segments,
                            )
                        # Break out of the while loop and move to the next element.
                        break
                    else:
                        # If we can't match an element, we should ascertain whether it's
                        # required. If so then fine, move on, but otherwise we should
                        # crash out without a match. We have not matched the sequence.
                        if elem.is_optional():
                            # This will crash us out of the while loop and move us
                            # onto the next matching element
                            break
                        else:
                            return MatchResult.from_unmatched(segments)

        # If we get to here, we've matched all of the elements (or skipped them)
        # but still have some segments left (or perhaps have precisely zero left).
        # In either case, we're golden. Return successfully, with any leftovers as
        # the unmatched elements. Meta all go at the end regardless of wny trailing
        # whitespace.

        return MatchResult(
            BaseSegment._position_segments(
                matched_segments.matched_segments + meta_pre_nc + meta_post_nc,
            ),
            unmatched_segments,
        )


class Bracketed(Sequence):
    """Match if a bracketed sequence, with content that matches one of the elements.

    Note that the contents of the Bracketed Expression are treated as an expected
    sequence.

    Changelog:
    - Post 0.3.2: Bracketed inherits from Sequence and anything within
      the the `Bracketed()` expression is treated as a sequence. For the
      content of the Brackets, we call the `match()` method of the sequence
      grammar.
    - Post 0.1.0: Bracketed was separate from sequence, and the content
      of the expression were treated as options (like OneOf).
    - Pre 0.1.0: Bracketed inherited from Sequence and simply added
      brackets to that sequence.
    """

    def __init__(self, *args, **kwargs):
        # Store the bracket type. NB: This is only
        # hydrated into segments at runtime.
        self.bracket_type = kwargs.pop("bracket_type", "round")
        self.bracket_pairs_set = kwargs.pop("bracket_pairs_set", "bracket_pairs")
        # Allow optional override for special bracket-like things
        self.start_bracket = kwargs.pop("start_bracket", None)
        self.end_bracket = kwargs.pop("end_bracket", None)
        super().__init__(*args, **kwargs)

    @cached_method_for_parse_context
    def simple(self, parse_context: ParseContext, crumbs=None) -> Optional[List[str]]:
        """Does this matcher support a uppercase hash matching route?

        Bracketed does this easily, we just look for the bracket.
        """
        start_bracket, _, _ = self.get_bracket_from_dialect(parse_context)
        return start_bracket.simple(parse_context=parse_context, crumbs=crumbs)

    def get_bracket_from_dialect(self, parse_context):
        """Rehydrate the bracket segments in question."""
        for bracket_type, start_ref, end_ref, persists in parse_context.dialect.sets(
            self.bracket_pairs_set
        ):
            if bracket_type == self.bracket_type:
                start_bracket = parse_context.dialect.ref(start_ref)
                end_bracket = parse_context.dialect.ref(end_ref)
                break
        else:  # pragma: no cover
            raise ValueError(
                "bracket_type {!r} not found in bracket_pairs of {!r} dialect.".format(
                    self.bracket_type, parse_context.dialect.name
                )
            )
        return start_bracket, end_bracket, persists

    @match_wrapper()
    @allow_ephemeral
    def match(
        self, segments: Tuple["BaseSegment", ...], parse_context: ParseContext
    ) -> MatchResult:
        """Match if a bracketed sequence, with content that matches one of the elements.

        1. work forwards to find the first bracket.
           If we find something other that whitespace, then fail out.
        2. Once we have the first bracket, we need to bracket count forward to find its
           partner.
        3. Assuming we find its partner then we try and match what goes between them
           using the match method of Sequence.
           If we match, great. If not, then we return an empty match.
           If we never find its partner then we return an empty match but should
           probably log a parsing warning, or error?

        """
        # Trim ends if allowed.
        if self.allow_gaps:
            pre_nc, seg_buff, post_nc = trim_non_code_segments(segments)
        else:
            seg_buff = segments  # pragma: no cover TODO?

        # Rehydrate the bracket segments in question.
        # bracket_persists controls whether we make a BracketedSegment or not.
        start_bracket, end_bracket, bracket_persists = self.get_bracket_from_dialect(
            parse_context
        )
        # Allow optional override for special bracket-like things
        start_bracket = self.start_bracket or start_bracket
        end_bracket = self.end_bracket or end_bracket

        # Are we dealing with a pre-existing BracketSegment?
        if seg_buff[0].is_type("bracketed"):
            # NOTE: We copy the original segment here because otherwise we will begin to
            # edit a _reference_ and not a copy - and that may lead to unused matches
            # leaking out. https://github.com/sqlfluff/sqlfluff/issues/3277
            seg: BracketedSegment = cast(BracketedSegment, seg_buff[0].copy())
            # Check it's of the right kind of bracket
            if not start_bracket.match(seg.start_bracket, parse_context):
                # Doesn't match - return no match
                return MatchResult.from_unmatched(segments)

            content_segs = seg.segments[len(seg.start_bracket) : -len(seg.end_bracket)]
            bracket_segment = seg
            trailing_segments = seg_buff[1:]
        # Otherwise try and match the segments directly.
        else:
            # Look for the first bracket
            with parse_context.deeper_match() as ctx:
                start_match = start_bracket.match(seg_buff, parse_context=ctx)
            if start_match:
                seg_buff = start_match.unmatched_segments
            else:
                # Can't find the opening bracket. No Match.
                return MatchResult.from_unmatched(segments)

            # Look for the closing bracket
            content_segs, end_match, _ = self._bracket_sensitive_look_ahead_match(
                segments=seg_buff,
                matchers=[end_bracket],
                parse_context=parse_context,
                start_bracket=start_bracket,
                end_bracket=end_bracket,
                bracket_pairs_set=self.bracket_pairs_set,
            )
            if not end_match:  # pragma: no cover
                raise SQLParseError(
                    "Couldn't find closing bracket for opening bracket.",
                    segment=start_match.matched_segments[0],
                )

            # Construct a bracket segment
            bracket_segment = BracketedSegment(
                segments=(
                    start_match.matched_segments
                    + content_segs
                    + end_match.matched_segments
                ),
                start_bracket=start_match.matched_segments,
                end_bracket=end_match.matched_segments,
            )
            trailing_segments = end_match.unmatched_segments

        # Then trim whitespace and deal with the case of non-code content e.g. "(   )"
        if self.allow_gaps:
            pre_segs, content_segs, post_segs = trim_non_code_segments(content_segs)
        else:  # pragma: no cover TODO?
            pre_segs = ()
            post_segs = ()

        # If we've got a case of empty brackets check whether that is allowed.
        if not content_segs:
            if not self._elements or (
                all(e.is_optional() for e in self._elements)
                and (self.allow_gaps or (not pre_segs and not post_segs))
            ):
                return MatchResult(
                    (bracket_segment,)
                    if bracket_persists
                    else bracket_segment.segments,
                    trailing_segments,
                )
            else:
                return MatchResult.from_unmatched(segments)

        # Match the content using super. Sequence will interpret the content of the
        # elements.
        with parse_context.deeper_match() as ctx:
            content_match = super().match(content_segs, parse_context=ctx)

        # We require a complete match for the content (hopefully for obvious reasons)
        if content_match.is_complete():
            # Reconstruct the bracket segment post match.
            # We need to realign the meta segments so the pos markers are correct.
            # Have we already got indents?
            meta_idx = None
            for idx, seg in enumerate(bracket_segment.segments):
                if (
                    seg.is_meta
                    and cast(MetaSegment, seg).indent_val > 0
                    and not cast(MetaSegment, seg).is_template
                ):
                    meta_idx = idx
                    break
            # If we've already got indents, don't add more.
            if meta_idx:
                bracket_segment.segments = BaseSegment._position_segments(
                    bracket_segment.start_bracket
                    + pre_segs
                    + content_match.all_segments()
                    + post_segs
                    + bracket_segment.end_bracket
                )
            # Append some indent and dedent tokens at the start and the end.
            else:
                bracket_segment.segments = BaseSegment._position_segments(
                    # NB: The nc segments go *outside* the indents.
                    bracket_segment.start_bracket
                    + (Indent(),)  # Add a meta indent here
                    + pre_segs
                    + content_match.all_segments()
                    + post_segs
                    + (Dedent(),)  # Add a meta indent here
                    + bracket_segment.end_bracket
                )
            return MatchResult(
                (bracket_segment,) if bracket_persists else bracket_segment.segments,
                trailing_segments,
            )
        # No complete match. Fail.
        else:
            return MatchResult.from_unmatched(segments)



"""A list of all Materialize SQL key words.

https://materialize.com/docs/sql/identifiers
"""

materialize_reserved_keywords = """ALL
ALTER
AND
ANY
AS
BY
CAST
CHECK
CLUSTER
CLUSTERS
CONNECTION
CONNECTIONS
CONSTRAINT
CREATE
CROSS
CURRENT
DELETE
DISTINCT
DROP
ELSE
EXISTS
FOLLOWING
FOR
FROM
FULL
GROUP
ILIKE
IN
INNER
INSERT
INTERSECT
INTO
IS
JOIN
LATERAL
LEFT
LIKE
NATURAL
NOT
NULL
NULLIF
OF
ON
OR
ORDER
ROW
ROWS
SELECT
SET
SINK
SINKS
TO
UNION
UNIQUE
UPDATE
USING
VALUES
WHEN
WHERE
WITH
"""

materialize_unreserved_keywords = """ACCESS
ACKS
ARN
ARRANGEMENT
ARRAY
ASC
AT
AUCTION
AUTHORITY
AVAILABILITY
AVRO
AWS
BEGIN
BETWEEN
BIGINT
BOOLEAN
BOTH
BPCHAR
BROKER
BROKERS
BUCKET
BYTES
CASCADE
CASE
CERTIFICATE
CHAIN
CHAR
CHARACTER
CHARACTERISTICS
CLIENT
CLOSE
COALESCE
COLLATE
COLUMNS
COMMIT
COMMITTED
COMPACTION
COMPRESSION
COMPUTE
CONFLUENT
COPY
COUNT
COUNTER
CSV
CURSOR
DATABASE
DATABASES
DATUMS
DAY
DAYS
DEALLOCATE
DEBEZIUM
DEBUG
DEBUGGING
DEC
DECIMAL
DECLARE
DECORRELATED
DEFAULT
DELIMITED
DELIMITER
DESC
DETAILS
DISCARD
DISCOVER
DOT
DOUBLE
ELEMENT
ENABLE
ENABLED
END
ENDPOINT
ENFORCED
ENVELOPE
ESCAPE
EXCEPT
EXECUTE
EXPECTED
EXPLAIN
EXTRACT
FACTOR
FALSE
FETCH
FIELDS
FILTER
FIRST
FLOAT
FOREIGN
FORMAT
FORWARD
FULLNAME
GENERATOR
GRAPH
GREATEST
GROUPS
GZIP
HAVING
HEADER
HEADERS
HOLD
HOST
HOUR
HOURS
ID
IDEMPOTENCE
IF
IGNORE
INCLUDE
INDEX
INDEXES
INFO
INLINE
INT
INTEGER
INTERSECT
INTERVAL
INTROSPECTION
ISNULL
ISOLATION
JSON
KAFKA
KEY
KEYS
KINESIS
LAST
LATEST
LEADING
LEAST
LEVEL
LIMIT
LIST
LOAD
LOCAL
LOG
LOGICAL
LOGIN
MAP
MATCHING
MATERIALIZE
MATERIALIZED
MAX
MECHANISMS
MESSAGE
METADATA
MINUTE
MINUTES
MODE
MONTH
MONTHS
MS
NAME
NAMES
NEXT
NO
NOLOGIN
NONE
NOSUPERUSER
NOTICE
NOTIFICATIONS
NULLS
OBJECTS
OFFSET
ONLY
OPERATOR
OPTIMIZED
OPTIMIZER
OPTIONS
ORDINALITY
OUTER
OVER
PARTITION
PASSWORD
PHYSICAL
PLAN
PLANS
PORT
POSITION
POSTGRES
PRECEDING
PRECISION
PREFIX
PREPARE
PRIMARY
PRIVATELINK
PROGRESS
PROTOBUF
PUBLICATION
QUERY
QUOTE
RAISE
RANGE
RAW
READ
REAL
REFERENCES
REFRESH
REGEX
REGION
REGISTRY
REMOTE
RENAME
REPEATABLE
REPLACE
REPLICA
REPLICAS
REPLICATION
RESET
RESTRICT
RETENTION
RETURNING
RIGHT
ROLE
ROLES
ROLLBACK
ROTATE
S3
SASL
SCALE
SCAN
SCHEMA
SCHEMAS
SCRIPT
SECOND
SECONDS
SECRET
SECRETS
SEED
SEQUENCES
SERIALIZABLE
SERVICE
SESSION
SHOW
SIZE
SMALLINT
SNAPSHOT
SOME
SOURCE
SOURCES
SQS
SSH
SSL
START
STDIN
STDOUT
STRATEGY
STRING
SUBSCRIBE
SUBSOURCE
SUBSTRING
SUPERUSER
SYSTEM
TABLE
TABLES
TAIL
TEMP
TEMPORARY
TEST
TEXT
THEN
TICK
TIES
TIME
TIMELINE
TIMEOUT
TIMESTAMP
TOKEN
TOPIC
TPCH
TRACE
TRAILING
TRANSACTION
TRIM
TRUE
TUNNEL
TYPE
TYPES
UNBOUNDED
UNCOMMITTED
UNKNOWN
UPSERT
URL
USER
USERNAME
USERS
VALUE
VARCHAR
VARYING
VIEW
VIEWS
WARNING
WINDOW
WIRE
WITHOUT
WORK
WORKERS
WRITE
YEAR
YEARS
ZONE
ZONES
"""



"""Common test fixtures for reflow modules."""

import pytest

from sqlfluff.core import FluffConfig


@pytest.fixture()
def default_config():
    """Return the default config for reflow tests."""
    return FluffConfig(overrides={"dialect": "ansi"})



"""Implementation of Rule L047."""
from typing import Optional

from sqlfluff.core.rules import BaseRule, LintFix, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler
from sqlfluff.core.rules.doc_decorators import (
    document_configuration,
    document_fix_compatible,
    document_groups,
)
from sqlfluff.utils.functional import sp, FunctionalContext


@document_groups
@document_fix_compatible
@document_configuration
class Rule_L047(BaseRule):
    """Use consistent syntax to express "count number of rows".

    Note:
        If both ``prefer_count_1`` and ``prefer_count_0`` are set to true
        then ``prefer_count_1`` has precedence.

    ``COUNT(*)``, ``COUNT(1)``, and even ``COUNT(0)`` are equivalent syntaxes
    in many SQL engines due to optimizers interpreting these instructions as
    "count number of rows in result".

    The ANSI-92_ spec mentions the ``COUNT(*)`` syntax specifically as
    having a special meaning:

        If COUNT(*) is specified, then
        the result is the cardinality of T.

    So by default, `SQLFluff` enforces the consistent use of ``COUNT(*)``.

    If the SQL engine you work with, or your team, prefers ``COUNT(1)`` or
    ``COUNT(0)`` over ``COUNT(*)``, you can configure this rule to consistently
    enforce your preference.

    .. _ANSI-92: http://msdn.microsoft.com/en-us/library/ms175997.aspx

    **Anti-pattern**

    .. code-block:: sql

        select
            count(1)
        from table_a

    **Best practice**

    Use ``count(*)`` unless specified otherwise by config ``prefer_count_1``,
    or ``prefer_count_0`` as preferred.

    .. code-block:: sql

        select
            count(*)
        from table_a

    """

    groups = ("all", "core")
    config_keywords = ["prefer_count_1", "prefer_count_0"]
    crawl_behaviour = SegmentSeekerCrawler({"function"})

    def _eval(self, context: RuleContext) -> Optional[LintResult]:
        """Find rule violations and provide fixes."""
        # Config type hints
        self.prefer_count_0: bool
        self.prefer_count_1: bool

        if (
            # We already know we're in a function because of the crawl_behaviour
            context.segment.get_child("function_name").raw_upper
            == "COUNT"
        ):
            # Get bracketed content
            f_content = (
                FunctionalContext(context)
                .segment.children(sp.is_type("bracketed"))
                .children(
                    sp.and_(
                        sp.not_(sp.is_meta()),
                        sp.not_(
                            sp.is_type(
                                "start_bracket", "end_bracket", "whitespace", "newline"
                            )
                        ),
                    )
                )
            )
            if len(f_content) != 1:  # pragma: no cover
                return None

            preferred = "*"
            if self.prefer_count_1:
                preferred = "1"
            elif self.prefer_count_0:
                preferred = "0"

            if f_content[0].is_type("star") and (
                self.prefer_count_1 or self.prefer_count_0
            ):
                return LintResult(
                    anchor=context.segment,
                    fixes=[
                        LintFix.replace(
                            f_content[0],
                            [
                                f_content[0].edit(
                                    f_content[0].raw.replace("*", preferred)
                                )
                            ],
                        ),
                    ],
                )

            if f_content[0].is_type("expression"):
                expression_content = [
                    seg for seg in f_content[0].segments if not seg.is_meta
                ]

                if (
                    len(expression_content) == 1
                    and expression_content[0].is_type("literal")
                    and expression_content[0].raw in ["0", "1"]
                    and expression_content[0].raw != preferred
                ):
                    return LintResult(
                        anchor=context.segment,
                        fixes=[
                            LintFix.replace(
                                expression_content[0],
                                [
                                    expression_content[0].edit(
                                        expression_content[0].raw.replace(
                                            expression_content[0].raw, preferred
                                        )
                                    ),
                                ],
                            ),
                        ],
                    )
        return None



"""Surrogate class for working with TemplatedFileSlice collections."""
from typing import Callable, Optional

from sqlfluff.core.templaters.base import TemplatedFileSlice, TemplatedFile


class TemplatedFileSlices(tuple):
    """Encapsulates a sequence of one or more TemplatedFileSlice.

    The slices may or may not be contiguous in a file.
    Provides useful operations on a sequence of slices to simplify rule creation.
    """

    def __new__(cls, *templated_slices, templated_file=None):
        """Override new operator."""
        return super(TemplatedFileSlices, cls).__new__(cls, templated_slices)

    def __init__(self, *_: TemplatedFileSlice, templated_file: TemplatedFile):
        self.templated_file = templated_file

    def all(
        self, predicate: Optional[Callable[[TemplatedFileSlice], bool]] = None
    ) -> bool:
        """Do all the templated slices match?"""
        for s in self:
            if predicate is not None and not predicate(s):
                return False
        return True

    def any(
        self, predicate: Optional[Callable[[TemplatedFileSlice], bool]] = None
    ) -> bool:  # pragma: no cover
        """Do any of the templated slices match?"""
        for s in self:
            if predicate is None or predicate(s):
                return True
        return False

    def select(
        self,
        select_if: Optional[Callable[[TemplatedFileSlice], bool]] = None,
        loop_while: Optional[Callable[[TemplatedFileSlice], bool]] = None,
        start_slice: Optional[TemplatedFileSlice] = None,
        stop_slice: Optional[TemplatedFileSlice] = None,
    ) -> "TemplatedFileSlices":  # pragma: no cover
        """Retrieve range/subset.

        NOTE: Iterates the slices BETWEEN start_slice and stop_slice, i.e. those
        slices are not included in the loop.
        """
        start_index = self.index(start_slice) if start_slice else -1
        stop_index = self.index(stop_slice) if stop_slice else len(self)
        buff = []
        for slice_ in self[start_index + 1 : stop_index]:
            if loop_while is not None and not loop_while(slice_):
                break
            if select_if is None or select_if(slice_):
                buff.append(slice_)
        return TemplatedFileSlices(*buff, templated_file=self.templated_file)



"""Implementation of Rule L028."""

from typing import Iterator, List, Optional, Set

from sqlfluff.core.dialects.common import AliasInfo, ColumnAliasInfo
from sqlfluff.core.parser.segments.base import BaseSegment, IdentitySet
from sqlfluff.core.parser.segments.raw import SymbolSegment
from sqlfluff.utils.analysis.select import SelectStatementColumnsAndTables
from sqlfluff.utils.analysis.select_crawler import Query, SelectCrawler
from sqlfluff.core.rules import (
    BaseRule,
    LintFix,
    LintResult,
    EvalResultType,
    RuleContext,
)
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler
from sqlfluff.utils.functional import sp, FunctionalContext
from sqlfluff.core.rules.doc_decorators import (
    document_configuration,
    document_fix_compatible,
    document_groups,
)
from sqlfluff.dialects.dialect_ansi import IdentifierSegment


_START_TYPES = ["select_statement", "set_expression", "with_compound_statement"]


@document_groups
@document_fix_compatible
@document_configuration
class Rule_L028(BaseRule):
    """References should be consistent in statements with a single table.

    .. note::
        For BigQuery, Hive and Redshift this rule is disabled by default.
        This is due to historical false positives associated with STRUCT data types.
        This default behaviour may be changed in the future.
        The rule can be enabled with the ``force_enable = True`` flag.

    "consistent" will be fixed to "qualified" if inconsistency is found.

    **Anti-pattern**

    In this example, only the field ``b`` is referenced.

    .. code-block:: sql

        SELECT
            a,
            foo.b
        FROM foo

    **Best practice**

    Add or remove references to all fields.

    .. code-block:: sql

        SELECT
            a,
            b
        FROM foo

        -- Also good

        SELECT
            foo.a,
            foo.b
        FROM foo

    """

    groups = ("all",)
    config_keywords = [
        "single_table_references",
        "force_enable",
    ]
    crawl_behaviour = SegmentSeekerCrawler(set(_START_TYPES))
    _is_struct_dialect = False
    _dialects_with_structs = ["bigquery", "hive", "redshift"]
    # This could be turned into an option
    _fix_inconsistent_to = "qualified"

    def _eval(self, context: RuleContext) -> EvalResultType:
        """Override base class for dialects that use structs, or SELECT aliases."""
        # Config type hints
        self.force_enable: bool
        # Some dialects use structs (e.g. column.field) which look like
        # table references and so incorrectly trigger this rule.
        if (
            context.dialect.name in self._dialects_with_structs
            and not self.force_enable
        ):
            return LintResult()

        if context.dialect.name in self._dialects_with_structs:
            self._is_struct_dialect = True

        if not FunctionalContext(context).parent_stack.any(sp.is_type(*_START_TYPES)):
            crawler = SelectCrawler(context.segment, context.dialect)
            visited: IdentitySet = IdentitySet()
            if crawler.query_tree:
                # Recursively visit and check each query in the tree.
                return list(self._visit_queries(crawler.query_tree, visited))
        return None

    def _visit_queries(
        self, query: Query, visited: IdentitySet
    ) -> Iterator[LintResult]:
        select_info: Optional[SelectStatementColumnsAndTables] = None
        if query.selectables:
            select_info = query.selectables[0].select_info
            # How many table names are visible from here? If more than one then do
            # nothing.
            if select_info and len(select_info.table_aliases) == 1:
                fixable = True
                # :TRICKY: Subqueries in the column list of a SELECT can see tables
                # in the FROM list of the containing query. Thus, count tables at
                # the *parent* query level.
                table_search_root = query.parent if query.parent else query
                query_list = (
                    SelectCrawler.get(
                        table_search_root, table_search_root.selectables[0].selectable
                    )
                    if table_search_root.selectables
                    else []
                )
                filtered_query_list = [q for q in query_list if isinstance(q, str)]
                if len(filtered_query_list) != 1:
                    # If more than one table name is visible, check for and report
                    # potential lint warnings, but don't generate fixes, because
                    # fixes are unsafe if there's more than one table visible.
                    fixable = False
                yield from _check_references(
                    select_info.table_aliases,
                    select_info.standalone_aliases,
                    select_info.reference_buffer,
                    select_info.col_aliases,
                    self.single_table_references,  # type: ignore
                    self._is_struct_dialect,
                    self._fix_inconsistent_to,
                    fixable,
                )
        children = list(query.children)
        # 'query.children' includes CTEs and "main" queries, but not queries in
        # the "FROM" list. We want to visit those as well.
        if select_info:
            for a in select_info.table_aliases:
                for q in SelectCrawler.get(query, a.from_expression_element):
                    if not isinstance(q, Query):
                        continue
                    # Check for previously visited selectables to avoid possible
                    # infinite recursion, e.g.:
                    #   WITH test1 AS (SELECT i + 1, j + 1 FROM test1)
                    #   SELECT * FROM test1;
                    if any(s.selectable in visited for s in q.selectables):
                        continue
                    visited.update(s.selectable for s in q.selectables)
                    children.append(q)
        for child in children:
            yield from self._visit_queries(child, visited)


def _check_references(
    table_aliases: List[AliasInfo],
    standalone_aliases: List[str],
    references: List[BaseSegment],
    col_aliases: List[ColumnAliasInfo],
    single_table_references: str,
    is_struct_dialect: bool,
    fix_inconsistent_to: Optional[str],
    fixable: bool,
) -> Iterator[LintResult]:
    """Iterate through references and check consistency."""
    # A buffer to keep any violations.
    col_alias_names: List[str] = [c.alias_identifier_name for c in col_aliases]
    table_ref_str: str = table_aliases[0].ref_str
    table_ref_str_source = table_aliases[0].segment
    # Check all the references that we have.
    seen_ref_types: Set[str] = set()
    for ref in references:
        this_ref_type: str = ref.qualification()  # type: ignore
        if this_ref_type == "qualified" and is_struct_dialect:
            # If this col appears "qualified" check if it is more logically a struct.
            if next(ref.iter_raw_references()).part != table_ref_str:  # type: ignore
                this_ref_type = "unqualified"

        lint_res = _validate_one_reference(
            single_table_references,
            ref,
            this_ref_type,
            standalone_aliases,
            table_ref_str,
            table_ref_str_source,
            col_alias_names,
            seen_ref_types,
            fixable,
        )

        seen_ref_types.add(this_ref_type)
        if not lint_res:
            continue

        if fix_inconsistent_to and single_table_references == "consistent":
            # If we found a "consistent" error but we have a fix directive,
            # recurse with a different single_table_references value
            yield from _check_references(
                table_aliases,
                standalone_aliases,
                references,
                col_aliases,
                # NB vars are passed in a different order here
                single_table_references=fix_inconsistent_to,
                is_struct_dialect=is_struct_dialect,
                fix_inconsistent_to=None,
                fixable=fixable,
            )

        yield lint_res


def _validate_one_reference(
    single_table_references: str,
    ref: BaseSegment,
    this_ref_type: str,
    standalone_aliases: List[str],
    table_ref_str: str,
    table_ref_str_source: Optional[BaseSegment],
    col_alias_names: List[str],
    seen_ref_types: Set[str],
    fixable: bool,
) -> Optional[LintResult]:
    # We skip any unqualified wildcard references (i.e. *). They shouldn't
    # count.
    if not ref.is_qualified() and ref.is_type("wildcard_identifier"):  # type: ignore
        return None
    # Oddball case: Column aliases provided via function calls in by
    # FROM or JOIN. References to these don't need to be qualified.
    # Note there could be a table with a column by the same name as
    # this alias, so avoid bogus warnings by just skipping them
    # entirely rather than trying to enforce anything.
    if ref.raw in standalone_aliases:
        return None

    # Oddball case: tsql table variables can't be used to qualify references.
    # This appears here as an empty string for table_ref_str.
    if not table_ref_str:
        return None

    # Certain dialects allow use of SELECT alias in WHERE clauses
    if ref.raw in col_alias_names:
        return None

    if single_table_references == "consistent":
        if seen_ref_types and this_ref_type not in seen_ref_types:
            return LintResult(
                anchor=ref,
                description=f"{this_ref_type.capitalize()} reference "
                f"{ref.raw!r} found in single table select which is "
                "inconsistent with previous references.",
            )

        return None

    if single_table_references != this_ref_type:
        if single_table_references == "unqualified":
            # If this is qualified we must have a "table", "."" at least
            fixes = [LintFix.delete(el) for el in ref.segments[:2]] if fixable else None
            return LintResult(
                anchor=ref,
                fixes=fixes,
                description="{} reference {!r} found in single table "
                "select.".format(this_ref_type.capitalize(), ref.raw),
            )

        fixes = None
        if fixable:
            fixes = [
                LintFix.create_before(
                    ref.segments[0] if len(ref.segments) else ref,
                    source=[table_ref_str_source] if table_ref_str_source else None,
                    edit_segments=[
                        IdentifierSegment(
                            raw=table_ref_str,
                            type="naked_identifier",
                        ),
                        SymbolSegment(raw=".", type="symbol"),
                    ],
                )
            ]
        return LintResult(
            anchor=ref,
            fixes=fixes,
            description="{} reference {!r} found in single table "
            "select.".format(this_ref_type.capitalize(), ref.raw),
        )

    return None



"""Raw segment definitions.

This is designed to be the root segment, without
any children, and the output of the lexer.
"""

from typing import List, Optional, Tuple, Set
from uuid import UUID, uuid4

from sqlfluff.core.parser.segments.base import BaseSegment, SourceFix
from sqlfluff.core.parser.markers import PositionMarker


class RawSegment(BaseSegment):
    """This is a segment without any subsegments."""

    type = "raw"
    _is_code = True
    _is_comment = False
    _is_whitespace = False
    # Classes inheriting from RawSegment may provide a _default_raw
    # to enable simple initialisation.
    _default_raw = ""

    def __init__(
        self,
        raw: Optional[str] = None,
        pos_marker: Optional[PositionMarker] = None,
        type: Optional[str] = None,
        trim_start: Optional[Tuple[str, ...]] = None,
        trim_chars: Optional[Tuple[str, ...]] = None,
        source_fixes: Optional[List[SourceFix]] = None,
        uuid: Optional[UUID] = None,
    ):
        """Initialise raw segment.

        If raw is not provided, we default to _default_raw if present.
        If pos_marker is not provided, it is assume that this will be
        inserted later as part of a reposition phase.
        """
        if raw is not None:  # NB, raw *can* be an empty string and be valid
            self._raw = raw
        else:
            self._raw = self._default_raw
        self._raw_upper = self._raw.upper()
        # pos marker is required here. We ignore the typing initially
        # because it might *initially* be unset, but it will be reset
        # later.
        self.pos_marker: PositionMarker = pos_marker  # type: ignore
        # if a surrogate type is provided, store it for later.
        self._surrogate_type = type
        # What should we trim off the ends to get to content
        self.trim_start = trim_start
        self.trim_chars = trim_chars
        # A cache variable for expandable
        self._is_expandable = None
        # Keep track of any source fixes
        self._source_fixes = source_fixes
        # UUID for matching
        self.uuid = uuid or uuid4()

    def __repr__(self):
        return "<{}: ({}) {!r}>".format(
            self.__class__.__name__, self.pos_marker, self.raw
        )

    def __setattr__(self, key, value):
        """Overwrite BaseSegment's __setattr__ with BaseSegment's superclass."""
        super(BaseSegment, self).__setattr__(key, value)

    # ################ PUBLIC PROPERTIES

    @property
    def matched_length(self):
        """Return the length of the segment in characters."""
        return len(self._raw)

    @property
    def is_expandable(self):
        """Return true if it is meaningful to call `expand` on this segment."""
        return False

    @property
    def is_code(self):
        """Return True if this segment is code."""
        return self._is_code

    @property
    def is_comment(self):
        """Return True if this segment is a comment."""
        return self._is_comment

    @property
    def is_whitespace(self):
        """Return True if this segment is whitespace."""
        return self._is_whitespace

    @property
    def raw(self):
        """Returns the raw segment."""
        return self._raw

    @property
    def raw_upper(self):
        """Returns the raw segment in uppercase."""
        return self._raw_upper

    @property
    def raw_segments(self):
        """Returns self to be compatible with calls to its superclass."""
        return [self]

    @property
    def segments(self):
        """Return an empty list of child segments.

        This is in case something tries to iterate on this segment.
        """
        return []

    @property
    def class_types(self) -> Set[str]:
        """The set of full types for this segment, including inherited.

        Add the surrogate type for raw segments.
        """
        return (
            {self._surrogate_type} if self._surrogate_type else set()
        ) | super().class_types

    @property
    def source_fixes(self) -> List[SourceFix]:
        """Return any source fixes as list."""
        return self._source_fixes or []

    # ################ INSTANCE METHODS

    def invalidate_caches(self):
        """Overwrite superclass functionality."""
        pass

    def get_type(self):
        """Returns the type of this segment as a string."""
        return self._surrogate_type or self.type

    def is_type(self, *seg_type):
        """Extend the parent class method with the surrogate types."""
        if self._surrogate_type and self._surrogate_type in seg_type:
            return True
        return self.class_is_type(*seg_type)

    def get_raw_segments(self):
        """Iterate raw segments, mostly for searching."""
        return [self]

    def raw_trimmed(self):
        """Return a trimmed version of the raw content."""
        raw_buff = self.raw
        if self.trim_start:
            for seq in self.trim_start:
                if raw_buff.startswith(seq):
                    raw_buff = raw_buff[len(seq) :]
        if self.trim_chars:
            raw_buff = self.raw
            # for each thing to trim
            for seq in self.trim_chars:
                # trim start
                while raw_buff.startswith(seq):
                    raw_buff = raw_buff[len(seq) :]
                # trim end
                while raw_buff.endswith(seq):
                    raw_buff = raw_buff[: -len(seq)]
            return raw_buff
        return raw_buff

    def raw_list(self):  # pragma: no cover TODO?
        """Return a list of the raw content of this segment."""
        return [self.raw]

    def stringify(self, ident=0, tabsize=4, code_only=False):
        """Use indentation to render this segment and its children as a string."""
        preface = self._preface(ident=ident, tabsize=tabsize)
        return preface + "\n"

    def _suffix(self):
        """Return any extra output required at the end when logging.

        NB Override this for specific subclasses if we want extra output.
        """
        return f"{self.raw!r}"

    def edit(
        self, raw: Optional[str] = None, source_fixes: Optional[List[SourceFix]] = None
    ):
        """Create a new segment, with exactly the same position but different content.

        Returns:
            A copy of this object with new contents.

        Used mostly by fixes.

        NOTE: This *doesn't* copy the uuid. The edited segment is a new segment.

        """
        return self.__class__(
            raw=raw or self.raw,
            pos_marker=self.pos_marker,
            type=self._surrogate_type,
            trim_start=self.trim_start,
            trim_chars=self.trim_chars,
            source_fixes=source_fixes or self.source_fixes,
        )


class CodeSegment(RawSegment):
    """An alias for RawSegment.

    This has a more explicit name for segment creation.
    """

    pass


class UnlexableSegment(CodeSegment):
    """A placeholder to unlexable sections.

    This otherwise behaves exactly like a code section.
    """

    type = "unlexable"


class CommentSegment(RawSegment):
    """Segment containing a comment."""

    type = "comment"
    _is_code = False
    _is_comment = True


class WhitespaceSegment(RawSegment):
    """Segment containing whitespace."""

    type = "whitespace"
    _is_whitespace = True
    _is_code = False
    _is_comment = False
    _default_raw = " "


class NewlineSegment(RawSegment):
    """Segment containing a newline.

    NOTE: NewlineSegment does not inherit from WhitespaceSegment.
    Therefore NewlineSegment.is_type('whitespace') returns False.

    This is intentional and convenient for rules. If users want
    to match on both, call .is_type('whitespace', 'newline')
    """

    type = "newline"
    _is_whitespace = True
    _is_code = False
    _is_comment = False
    _default_raw = "\n"


class KeywordSegment(CodeSegment):
    """A segment used for matching single words.

    We rename the segment class here so that descendants of
    _ProtoKeywordSegment can use the same functionality
    but don't end up being labelled as a `keyword` later.
    """

    type = "keyword"

    def __init__(
        self,
        raw: Optional[str] = None,
        pos_marker: Optional[PositionMarker] = None,
        type: Optional[str] = None,
        source_fixes: Optional[List[SourceFix]] = None,
    ):
        """If no other name is provided we extrapolate it from the raw."""
        super().__init__(
            raw=raw,
            pos_marker=pos_marker,
            type=type,
            source_fixes=source_fixes,
        )

    def edit(self, raw=None, source_fixes=None):
        """Create a new segment, with exactly the same position but different content.

        Returns:
            A copy of this object with new contents.

        Used mostly by fixes.

        NOTE: This *doesn't* copy the uuid. The edited segment is a new segment.

        """
        return self.__class__(
            raw=raw or self.raw,
            pos_marker=self.pos_marker,
            type=self._surrogate_type,
            source_fixes=source_fixes or self.source_fixes,
        )


class SymbolSegment(CodeSegment):
    """A segment used for matching single entities which aren't keywords.

    We rename the segment class here so that descendants of
    _ProtoKeywordSegment can use the same functionality
    but don't end up being labelled as a `keyword` later.
    """

    type = "symbol"



"""Implementation of Rule L038."""
from typing import Optional

from sqlfluff.core.parser import BaseSegment, SymbolSegment

from sqlfluff.core.rules import BaseRule, LintFix, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler
from sqlfluff.utils.functional import sp, FunctionalContext
from sqlfluff.core.rules.doc_decorators import (
    document_configuration,
    document_fix_compatible,
    document_groups,
)


@document_groups
@document_fix_compatible
@document_configuration
class Rule_L038(BaseRule):
    """Trailing commas within select clause.

    .. note::
       For many database backends this is allowed. For some users
       this may be something they wish to enforce (in line with
       Python best practice). Many database backends regard this
       as a syntax error, and as such the `SQLFluff` default is to
       forbid trailing commas in the select clause.

    **Anti-pattern**

    .. code-block:: sql

        SELECT
            a,
            b,
        FROM foo

    **Best practice**

    .. code-block:: sql

        SELECT
            a,
            b
        FROM foo
    """

    groups = ("all", "core")
    config_keywords = ["select_clause_trailing_comma"]
    crawl_behaviour = SegmentSeekerCrawler({"select_clause"})

    def _eval(self, context: RuleContext) -> Optional[LintResult]:
        """Trailing commas within select clause."""
        # Config type hints
        self.select_clause_trailing_comma: str

        segment = FunctionalContext(context).segment
        children = segment.children()
        # Iterate content to find last element
        last_content: BaseSegment = children.last(sp.is_code())[0]

        # What mode are we in?
        if self.select_clause_trailing_comma == "forbid":
            # Is it a comma?
            if last_content.is_type("comma"):
                return LintResult(
                    anchor=last_content,
                    fixes=[LintFix.delete(last_content)],
                    description="Trailing comma in select statement forbidden",
                )
        elif self.select_clause_trailing_comma == "require":
            if not last_content.is_type("comma"):
                new_comma = SymbolSegment(",", type="comma")
                return LintResult(
                    anchor=last_content,
                    fixes=[LintFix.replace(last_content, [last_content, new_comma])],
                    description="Trailing comma in select statement required",
                )
        return None



"""Classes to help with match logging."""

from sqlfluff.core.parser.helpers import join_segments_raw_curtailed


class LateLoggingObject:
    """A basic late binding log object for parse_match_logging.

    This allows us to defer the string manipulation involved
    until actually required by the logger.
    """

    __slots__ = "v_level", "logger", "msg"

    def __init__(self, logger, msg, v_level=3):
        self.v_level = v_level
        self.logger = logger
        self.msg = msg

    def __str__(self):  # pragma: no cover TODO?
        """Actually materialise the string."""
        return self.msg

    def log(self):
        """Actually log this object."""
        # Otherwise carry on...
        if self.v_level == 3:
            self.logger.info(self)
        elif self.v_level == 4:
            self.logger.debug(self)


class ParseMatchLogObject(LateLoggingObject):
    """A late binding log object for parse_match_logging.

    This allows us to defer the string manipulation involved
    until actually required by the logger.
    """

    __slots__ = [
        "context",
        "grammar",
        "func",
        "kwargs",
    ]

    def __init__(self, parse_context, grammar, func, msg, v_level=3, **kwargs):
        super().__init__(v_level=v_level, logger=parse_context.logger, msg=msg)
        self.context = parse_context
        self.grammar = grammar
        self.func = func
        self.kwargs = kwargs

    def __str__(self):
        """Actually materialise the string."""
        symbol = self.kwargs.pop("symbol", "")
        s = "[PD:{:<2} MD:{:<2}]\t{:<50}\t{:<20}\t{:<4}".format(
            self.context.parse_depth,
            self.context.match_depth,
            ("." * self.context.match_depth) + str(self.context.match_segment),
            f"{self.grammar:.5}.{self.func} {self.msg}",
            symbol,
        )
        if self.kwargs:
            s += "\t[{}]".format(
                ", ".join(
                    f"{k}={repr(v) if isinstance(v, str) else str(v)}"
                    for k, v in self.kwargs.items()
                )
            )
        return s


def parse_match_logging(grammar, func, msg, parse_context, v_level=3, **kwargs):
    """Log in a particular consistent format for use while matching."""
    # Make a late bound log object so we only do the string manipulation when we need
    # to.
    ParseMatchLogObject(
        parse_context, grammar, func, msg, v_level=v_level, **kwargs
    ).log()


class LateBoundJoinSegmentsCurtailed:
    """Object to delay `join_segments_raw_curtailed` until later.

    This allows us to defer the string manipulation involved
    until actually required by the logger.
    """

    def __init__(self, segments):
        self.segments = segments

    def __str__(self):
        return repr(join_segments_raw_curtailed(self.segments))



"""Defines the Parser class."""

from typing import Optional, Sequence, TYPE_CHECKING

from sqlfluff.core.parser.context import RootParseContext
from sqlfluff.core.config import FluffConfig

if TYPE_CHECKING:
    from sqlfluff.core.parser.segments import BaseSegment  # pragma: no cover


class Parser:
    """Instantiates parsed queries from a sequence of lexed raw segments."""

    def __init__(
        self, config: Optional[FluffConfig] = None, dialect: Optional[str] = None
    ):
        # Allow optional config and dialect
        self.config = FluffConfig.from_kwargs(config=config, dialect=dialect)
        self.RootSegment = self.config.get("dialect_obj").get_root_segment()

    def parse(
        self,
        segments: Sequence["BaseSegment"],
        recurse=True,
        fname: Optional[str] = None,
    ) -> Optional["BaseSegment"]:
        """Parse a series of lexed tokens using the current dialect."""
        if not segments:  # pragma: no cover
            # This should normally never happen because there will usually
            # be an end_of_file segment. It would probably only happen in
            # api use cases.
            return None
        # Instantiate the root segment
        root_segment = self.RootSegment(segments=segments, fname=fname)
        # Call .parse() on that segment

        with RootParseContext.from_config(config=self.config, recurse=recurse) as ctx:
            parsed = root_segment.parse(parse_context=ctx)

        return parsed



"""This is an example of how to use the simple sqlfluff api."""

from typing import Any, Dict, Iterator, List, Union
import sqlfluff

#  -------- LINTING ----------

my_bad_query = "SeLEct  *, 1, blah as  fOO  from mySchema.myTable"

# Lint the given string and return an array of violations in JSON representation.
lint_result = sqlfluff.lint(my_bad_query, dialect="bigquery")
# lint_result =
# [
#     {
#         "code": "L010",
#         "line_no": 1,
#         "line_pos": 1,
#         "description": "Keywords must be consistently upper case.",
#     }
#     ...
# ]

#  -------- FIXING ----------

# Fix the given string and get a string back which has been fixed.
fix_result_1 = sqlfluff.fix(my_bad_query, dialect="bigquery")
# fix_result_1 = 'SELECT  *, 1, blah AS  foo  FROM myschema.mytable\n'

# We can also fix just specific rules.
fix_result_2 = sqlfluff.fix(my_bad_query, rules=["L010"])
# fix_result_2 = 'SELECT  *, 1, blah AS  fOO  FROM mySchema.myTable'

# Or a subset of rules...
fix_result_3 = sqlfluff.fix(my_bad_query, rules=["L010", "L014"])
# fix_result_3 = 'SELECT  *, 1, blah AS  fOO  FROM myschema.mytable'

#  -------- PARSING ----------

# Parse the given string and return a JSON representation of the parsed tree.
parse_result = sqlfluff.parse(my_bad_query)
# parse_result = {'file': {'statement': {...}, 'newline': '\n'}}

# This JSON structure can then be parsed as required.
# An example usage is shown below:


def get_json_segment(
    parse_result: Dict[str, Any], segment_type: str
) -> Iterator[Union[str, Dict[str, Any], List[Dict[str, Any]]]]:
    """Recursively search JSON parse result for specified segment type.

    Args:
        parse_result (Dict[str, Any]): JSON parse result from `sqlfluff.fix`.
        segment_type (str): The segment type to search for.

    Yields:
        Iterator[Union[str, Dict[str, Any], List[Dict[str, Any]]]]:
        Retrieves children of specified segment type as either a string for a raw
        segment or as JSON or an array of JSON for non-raw segments.
    """
    for k, v in parse_result.items():
        if k == segment_type:
            yield v
        elif isinstance(v, dict):
            yield from get_json_segment(v, segment_type)
        elif isinstance(v, list):
            for s in v:
                yield from get_json_segment(s, segment_type)


# e.g. Retrieve array of JSON for table references.
table_references = list(get_json_segment(parse_result, "table_reference"))
print(table_references)
# [[{'identifier': 'mySchema'}, {'dot': '.'}, {'identifier': 'myTable'}]]

# Retrieve raw table name from last identifier in the table reference.
for table_reference in table_references:
    table_name = list(get_json_segment(parse_result, "naked_identifier"))[-1]
    print(f"table_name: {table_name}")
# table_name: myTable
